from collections import defaultdict
from dataclasses import dataclass
import pytest

import toloka.client as toloka

from crowdom import base, mapping
from crowdom.objects import Audio, Image, Text
from . import lib


class TestValues:
    @dataclass
    class TestObject(base.Object):
        name: str
        count: int
        ready: bool

    @dataclass
    class TestObjectChild(TestObject):
        desc: str

    @dataclass
    class InvalidObject1(base.Object):
        volume: float

    @dataclass
    class InvalidObject2(base.Object):
        audio_class: lib.AudioClass

    def test_primitive_types_conversion(self):
        obj = self.TestObject(name='my obj name', count=42, ready=True)
        values = {'name': 'my obj name', 'count': 42, 'ready': True}

        assert mapping.obj_to_values(obj) == values
        assert mapping.values_to_obj(values, self.TestObject) == obj

    def test_class_conversion(self):
        obj = lib.sp
        values = {base.CLASS_OBJ_FIELD: 'sp'}

        assert mapping.obj_to_values(obj) == values
        assert mapping.values_to_obj(values, lib.AudioClass) == obj

    def test_inherited_fields(self):
        obj = self.TestObjectChild(name='my obj name', count=42, ready=True, desc='detailed desc')
        values = {'name': 'my obj name', 'count': 42, 'ready': True, 'desc': 'detailed desc'}

        assert mapping.obj_to_values(obj) == values
        assert mapping.values_to_obj(values, self.TestObjectChild) == obj

    def test_invalid_objects(self):
        with pytest.raises(ValueError) as e:
            mapping.obj_to_values(self.InvalidObject1(volume=0.5))
        assert str(e.value) == "object field \"volume\" has unsupported type <class 'float'>"

        with pytest.raises(ValueError) as e:
            mapping.obj_to_values(self.InvalidObject2(audio_class=lib.sp))
        assert str(e.value) == "object field \"audio_class\" has unsupported type <enum 'AudioClass'>"

    def test_invalid_values(self):
        with pytest.raises(ValueError) as e:
            mapping.values_to_obj({base.CLASS_OBJ_FIELD: 1}, lib.AudioClass)
        assert str(e.value) == '1 is not a valid AudioClass'

        with pytest.raises(AssertionError) as e:
            mapping.values_to_obj({'name': 'my obj name', 'count': 'some', 'ready': True}, self.TestObject)
        assert str(e.value) == "obj field type <class 'int'> does not match to values type <class 'str'>"

        with pytest.raises(TypeError) as e:
            mapping.values_to_obj({'name': 'my obj name', 'count': 42}, self.TestObject)
        assert str(e.value).endswith("__init__() missing 1 required positional argument: 'ready'")


class TestMapping:
    def test_object_mapping(self):
        assert lib.audio_mapping.to_values(Audio(url='https://1.wav')) == {'audio_link': 'https://1.wav'}
        assert lib.audio_mapping.from_values({'audio_link': 'https://1.wav'}) == Audio(url='https://1.wav')

    def test_object_mapping_on_none(self):
        with pytest.raises(AssertionError) as e:
            lib.audio_mapping.to_values(None)
        assert str(e.value) == (
            "object ObjectMeta(type=<class 'crowdom.objects.std.Audio'>, name=None, "
            "title=None, required=True) is required"
        )

        with pytest.raises(AssertionError) as e:
            lib.audio_mapping.from_values({})
        assert str(e.value) == (
            "missing fields {'audio_link'} in required object ObjectMeta(type=<class "
            "'crowdom.objects.std.Audio'>, name=None, title=None, required=True)"
        )

        assert lib.text_mapping.from_values({}) is None
        assert lib.text_mapping.to_values(None) == {}

    def test_invalid_object_mapping_not_dataclass(self):
        class NotDataclass:
            pass

        with pytest.raises(AssertionError) as e:
            mapping.ObjectMapping(obj_meta=base.ObjectMeta(NotDataclass), obj_task_fields=()).validate()
        assert str(e.value) == f'{NotDataclass} is not a dataclass nor an options class'

    def test_invalid_object_mapping_missing_object_field(self):
        with pytest.raises(AssertionError) as e:
            mapping.ObjectMapping(obj_meta=base.ObjectMeta(Audio), obj_task_fields=(('link', 'url'),)).validate()
        assert str(e.value) == f'{Audio} does not have attribute "link"'

    def test_invalid_task_mapping_fields_intersection(self):
        with pytest.raises(AssertionError) as e:
            mapping.TaskMapping(
                input_mapping=(
                    mapping.ObjectMapping(obj_meta=base.ObjectMeta(Image), obj_task_fields=(('url', 'url'),)),
                    mapping.ObjectMapping(obj_meta=base.ObjectMeta(Image), obj_task_fields=(('url', 'url'),)),
                ),
                output_mapping=(
                    mapping.ObjectMapping(
                        obj_meta=base.ObjectMeta(lib.ImageClass), obj_task_fields=(('value', 'choice'),)
                    ),
                ),
            ).validate()
        assert str(e.value) == "fields intersection in mapping list: ['url', 'url']"

    def test_object_mapping_to_values_wrong_object_class(self):
        with pytest.raises(AssertionError) as e:
            lib.audio_mapping.to_values(Text(text='привет'))
        assert str(e.value) == f'passed {Text} does not correspond to mapping {Audio}'

    def test_task_mapping(self):
        assert lib.audio_transcript_ext_mapping.to_task((Audio(url='https://1.wav'),)) == toloka.Task(
            input_values={'audio_link': 'https://1.wav', 'id': 'Audio(url=\'https://1.wav\')'}
        )

        assert lib.audio_transcript_ext_mapping.from_task(
            toloka.Task(input_values={'audio_link': 'https://1.wav', 'id': 'Audio(url=\'https://1.wav\')'})
        ) == (Audio(url='https://1.wav'),)

        assert lib.audio_transcript_ext_mapping.to_solution((Text(text='привет'), lib.sp)) == toloka.solution.Solution(
            output_values={'output': 'привет', 'choice': 'sp'}
        )

        # text is optional
        assert lib.audio_transcript_ext_mapping.to_solution((None, lib.sp)) == toloka.solution.Solution(
            output_values={'choice': 'sp'}
        )

        assert lib.audio_transcript_ext_mapping.from_solution(
            toloka.solution.Solution(output_values={'output': 'привет', 'choice': 'sp'})
        ) == (Text(text='привет'), lib.sp)

        # text is optional
        assert lib.audio_transcript_ext_mapping.from_solution(
            toloka.solution.Solution(output_values={'choice': 'sp'})
        ) == (None, lib.sp)

    def test_task_mapping_input_objects_validation(self):
        with pytest.raises(AssertionError) as e:
            lib.audio_transcript_ext_mapping.to_solution([Text(text='привет')])
        assert str(e.value) == 'expected length 2, got 1'

        with pytest.raises(AssertionError) as e:
            lib.audio_transcript_ext_mapping.to_task([Text(text='привет')])
        assert (
            str(e.value) == "expected type <class 'crowdom.objects.std.Audio'>, got <class 'crowdom.objects.std.Text'>"
        )

        # audio class is required
        with pytest.raises(AssertionError) as e:
            lib.audio_transcript_ext_mapping.to_solution([Text(text='привет'), None])
        assert str(e.value) == "expected type <enum 'AudioClass'>, got <class 'NoneType'>"

    def test_control_task_mapping(self):
        assert lib.audio_transcript_ext_mapping.to_control_task(
            ((Audio(url='https://1.wav'),), (Text(text='привет'), lib.sp))
        ) == toloka.Task(
            input_values={'audio_link': 'https://1.wav', 'id': 'Audio(url=\'https://1.wav\')'},
            known_solutions=[
                toloka.Task.KnownSolution(output_values={'output': 'привет', 'choice': 'sp'}, correctness_weight=1)
            ],
            infinite_overlap=True,
        )

    def test_control_task_mapping_output_objects_validation(self):
        with pytest.raises(AssertionError):
            lib.audio_transcript_ext_mapping.to_control_task(((Audio(url='https://1.wav'),), (Text(text='привет'),)))

    def test_task_id(self):
        counter = defaultdict(int)
        for audio_text_pair in [
            (Audio(url='https://1.wav'), Text(text='привет')),
            (Audio(url='https://1.wav'), Text(text='')),
            (Audio(url='https://1.wav'), Text(text='привет')),
            (Audio(url='https://2.wav'), Text(text='привет')),
        ]:
            counter[lib.audio_transcript_check_mapping.task_id(audio_text_pair)] += 1

        assert {task_id.id: count for task_id, count in counter.items()} == {
            "Audio(url='https://1.wav') Text(text='привет')": 2,
            "Audio(url='https://1.wav') Text(text='')": 1,
            "Audio(url='https://2.wav') Text(text='привет')": 1,
        }

    def test_get_solutions(self):
        audios = [Audio(url=f'https://{i + 1}.wav') for i in range(3)]
        control_audio = Audio(url='https://42.wav')

        tasks = [lib.audio_transcript_ext_mapping.to_task((audio,)) for audio in audios]
        control_task = lib.audio_transcript_ext_mapping.to_control_task(
            ((control_audio,), (Text(text='пусть'), lib.sp))
        )

        assignment_1 = toloka.Assignment(
            tasks=[tasks[0], tasks[2], control_task],
            solutions=[
                lib.audio_transcript_ext_mapping.to_solution((Text(text='привет'), lib.sp)),
                lib.audio_transcript_ext_mapping.to_solution((Text(text='алло'), lib.sp)),
                lib.audio_transcript_ext_mapping.to_solution((Text(text='путь'), lib.sp)),
            ],
        )
        assignment_2 = toloka.Assignment(
            tasks=[tasks[2], control_task, tasks[1]],
            solutions=[
                lib.audio_transcript_ext_mapping.to_solution((Text(text=''), lib.sil)),
                lib.audio_transcript_ext_mapping.to_solution((Text(text='пить'), lib.sp)),
                lib.audio_transcript_ext_mapping.to_solution((Text(text='нет не надо'), lib.sp)),
            ],
        )
        assignments = [assignment_1, assignment_2]

        pool_input_objects = [(audio,) for audio in audios]

        for with_control_tasks, expected_assignments_solutions in (
            (
                False,
                [
                    (
                        assignment_1,
                        [
                            ((audios[0],), (Text(text='привет'), lib.sp)),
                            ((audios[2],), (Text(text='алло'), lib.sp)),
                        ],
                    ),
                    (
                        assignment_2,
                        [
                            ((audios[2],), (Text(text=''), lib.sil)),
                            ((audios[1],), (Text(text='нет не надо'), lib.sp)),
                        ],
                    ),
                ],
            ),
            (
                True,
                [
                    (
                        assignment_1,
                        [
                            ((audios[0],), (Text(text='привет'), lib.sp)),
                            ((audios[2],), (Text(text='алло'), lib.sp)),
                            ((control_audio,), (Text(text='путь'), lib.sp)),
                        ],
                    ),
                    (
                        assignment_2,
                        [
                            ((audios[2],), (Text(text=''), lib.sil)),
                            ((control_audio,), (Text(text='пить'), lib.sp)),
                            ((audios[1],), (Text(text='нет не надо'), lib.sp)),
                        ],
                    ),
                ],
            ),
        ):
            assert (
                mapping.get_assignments_solutions(assignments, lib.audio_transcript_ext_mapping, with_control_tasks)
                == expected_assignments_solutions
            )

        solutions = mapping.get_solutions(assignments, lib.audio_transcript_ext_mapping, pool_input_objects)
        solutions = [
            ([id(obj) for obj in input_objects], output_objects)  # input objects must be the same
            for input_objects, output_objects in solutions
        ]

        assert solutions == [
            (
                [id(audios[0])],
                [((Text(text='привет'), lib.sp), assignment_1)],
            ),
            (
                [id(audios[1])],
                [((Text(text='нет не надо'), lib.sp), assignment_2)],
            ),
            (
                [id(audios[2])],
                [
                    ((Text(text='алло'), lib.sp), assignment_1),
                    ((Text(text=''), lib.sil), assignment_2),
                ],
            ),
        ]

    def test_feedback_loop_mapping(self):
        assert mapping.generate_feedback_loop_mapping(
            input_objects_mapping=(lib.audio_mapping,),
            markup_objects_mapping=(lib.text_mapping, lib.audio_class_mapping),
            check_objects_mapping=(lib.binary_evaluation_mapping, lib.audio_class_mapping),
        ) == (
            mapping.TaskMapping(
                input_mapping=(lib.audio_mapping,), output_mapping=(lib.text_mapping, lib.audio_class_mapping)
            ),
            mapping.TaskMapping(
                input_mapping=(lib.audio_mapping, lib.text_mapping, lib.audio_class_mapping),
                output_mapping=(lib.binary_evaluation_mapping, lib.audio_class_mapping),
            ),
        )

    def test_feedback_loop_first_check_mapping_is_not_evaluation(self):
        with pytest.raises(AssertionError):
            mapping.generate_feedback_loop_mapping(
                input_objects_mapping=(lib.audio_mapping,),
                markup_objects_mapping=(lib.text_mapping, lib.audio_class_mapping),
                check_objects_mapping=(lib.audio_class_mapping, lib.binary_evaluation_mapping),
            )

    def test_toloka_spec_gen(self):
        task_mapping = mapping.TaskMapping(
            input_mapping=(lib.text_mapping, lib.image_mapping, lib.metadata_mapping),
            output_mapping=(
                lib.binary_evaluation_mapping,
                lib.image_class_mapping,
                lib.mos_mapping,
                lib.score_mapping,
                lib.comment_mapping,
            ),
        )
        toloka_input_spec = {
            'output': toloka.project.StringSpec(required=False),
            'image_url': toloka.project.UrlSpec(),
            'id': toloka.project.StringSpec(hidden=True),
            'extra': toloka.project.StringSpec(hidden=True),
        }
        toloka_output_spec = {
            'ok': toloka.project.BooleanSpec(allowed_values=[False, True]),
            'score': toloka.project.IntegerSpec(),
            'mos': toloka.project.FloatSpec(),
            'choice': toloka.project.StringSpec(allowed_values=['cat', 'crow', 'dog']),
            '_comment': toloka.project.StringSpec(required=False),
        }
        assert task_mapping.to_toloka_input_spec() == toloka_input_spec
        assert task_mapping.to_toloka_output_spec() == toloka_output_spec

    def test_toloka_spec_overlapping_fields(self):
        task_mapping = mapping.TaskMapping(
            input_mapping=(lib.text_mapping, lib.text_mapping), output_mapping=(lib.binary_evaluation_mapping,)
        )
        with pytest.raises(AssertionError):
            task_mapping.to_toloka_input_spec()


class TestProjectSuitability:
    def test_is_suitable(self):
        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.UrlSpec(),
                            'id': toloka.project.StringSpec(hidden=True),
                        },
                        output_spec={
                            'output': toloka.project.StringSpec(required=False),
                            'choice': toloka.project.StringSpec(allowed_values=['sil', 'sp']),
                            'comment': toloka.project.StringSpec(),  # not present in task mapping
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is True
        )

    def test_project_spec_missing_fields(self):
        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.UrlSpec(),
                            'id': toloka.project.StringSpec(hidden=True),
                        },
                        output_spec={
                            'choice': toloka.project.StringSpec(),
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is False
        )

        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.UrlSpec(),
                        },
                        output_spec={
                            'output': toloka.project.StringSpec(),
                            'choice': toloka.project.StringSpec(allowed_values=['sil', 'sp']),
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is False
        )

    def test_project_spec_not_matching_fields(self):
        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.IntegerSpec(),
                            'id': toloka.project.StringSpec(hidden=True),
                        },
                        output_spec={
                            'output': toloka.project.StringSpec(),
                            'choice': toloka.project.StringSpec(allowed_values=['sil', 'sp']),
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is False
        )

        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.StringSpec(),
                            'id': toloka.project.StringSpec(hidden=True),
                        },
                        output_spec={
                            'output': toloka.project.StringSpec(),
                            'choice': toloka.project.BooleanSpec(),
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is False
        )

        assert (
            mapping.check_project_is_suitable(
                project=toloka.Project(
                    task_spec=toloka.project.TaskSpec(
                        input_spec={
                            'audio_link': toloka.project.UrlSpec(),
                            'id': toloka.project.StringSpec(hidden=True),
                        },
                        output_spec={
                            'output': toloka.project.StringSpec(),
                            'choice': toloka.project.StringSpec(allowed_values=['sil', 'sp', 'new']),
                        },
                    )
                ),
                task_mapping=lib.audio_transcript_ext_mapping,
            )
            is False
        )



import datetime
from mock import patch, Mock

import toloka.client as toloka

from crowdom import pool as pool_params, control, worker
from crowdom.worker import human


@patch.object(pool_params, 'datetime', Mock(wraps=datetime))
def test_default_markup_pool_builder():
    pool_params.datetime.datetime.utcnow.return_value = datetime.datetime(2020, 11, 5)
    project_id = '0001'
    private_name = 'default markup config'
    public_description = 'dynamic pricing'
    reward_per_assignment = 0.03
    real_tasks_count = 10
    task_duration_hint = datetime.timedelta(seconds=12)
    assignment_duration = real_tasks_count * task_duration_hint
    task_duration_ratio_rand, task_duration_ratio_poor = 0.1, 0.3
    markup_config = pool_params.MarkupConfig(
        project_id=project_id,
        private_name=private_name,
        public_description=public_description,
        reward_per_assignment=reward_per_assignment,
        task_duration_hint=task_duration_hint,
        real_tasks_count=real_tasks_count,
        control_params=control.Control(
            rules=control.RuleBuilder()
            .add_static_reward(0.5)
            .add_speed_control(task_duration_ratio_rand, task_duration_ratio_poor)
            .build()
        ),
    )
    actual_pool = pool_params.create_pool_params(markup_config)
    expected_pool = toloka.Pool(
        project_id=project_id,
        private_name=private_name,
        public_description=public_description,
        may_contain_adult_content=True,
        will_expire=datetime.datetime(2020, 12, 5),
        reward_per_assignment=reward_per_assignment,
        assignment_max_duration_seconds=int(assignment_duration.total_seconds()) * 5,
        auto_accept_solutions=False,
        auto_accept_period_day=14,
        assignments_issuing_config=toloka.Pool.AssignmentsIssuingConfig(issue_task_suites_in_creation_order=False),
        priority=30,
        defaults=toloka.Pool.Defaults(default_overlap_for_new_tasks=1, default_overlap_for_new_task_suites=1),
    )

    expected_pool.set_mixer_config(real_tasks_count=real_tasks_count, golden_tasks_count=0, training_tasks_count=0)

    expected_pool.quality_control.add_action(
        collector=toloka.collectors.SkippedInRowAssignments(),
        conditions=[toloka.conditions.SkippedInRowCount >= 5],
        action=toloka.actions.RestrictionV2(
            private_comment='Skipped assignments',
            duration=8,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_rand * markup_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, <= 0.1',
            duration=1 * 60 * 24,  # 1 day
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_poor * markup_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, 0.1 < time <= 0.3',
            duration=8 * 60,  # 8 hours
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AnswerCount(),
        conditions=[
            toloka.conditions.AssignmentsAcceptedCount >= int(datetime.timedelta(hours=1) / assignment_duration)
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Completed many assignments, vacation',
            duration=5,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.set_speed_quality_balance(toloka.pool.speed_quality_balance_config.TopPercentageByQuality(percent=90))
    assert actual_pool == expected_pool


@patch.object(human, 'datetime', Mock(wraps=datetime))
@patch.object(pool_params, 'datetime', Mock(wraps=datetime))
def test_markup_pool_builder():
    human.datetime.datetime.today.return_value = datetime.datetime(2020, 11, 5)
    pool_params.datetime.datetime.utcnow.return_value = datetime.datetime(2020, 11, 5)
    project_id = '0002'
    private_name = 'custom markup config'
    reward_per_assignment = 0.10
    real_tasks_count = 6
    task_duration_hint = datetime.timedelta(seconds=25)
    assignment_duration = real_tasks_count * task_duration_hint
    auto_accept_period_day = 3
    task_duration_ratio_rand, task_duration_ratio_poor = 0.2, 0.5
    training_requirement = toloka.pool.QualityControl.TrainingRequirement(
        training_pool_id='training pool', training_passing_skill_value=70
    )
    priority = 50
    markup_config = pool_params.MarkupConfig(
        project_id='0002',
        private_name=private_name,
        reward_per_assignment=0.10,
        task_duration_hint=task_duration_hint,
        real_tasks_count=6,
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='RU', verified=True)},
                    age_range=(10, 30),
                )
            ],
            training_score=70,
        ),
        training_requirement=training_requirement,
        priority=priority,
        ttl_days=20,
        auto_accept_period_day=auto_accept_period_day,
        may_contain_adult_content=False,
        control_params=control.Control(
            rules=control.RuleBuilder()
            .add_static_reward(0.5)
            .add_speed_control(task_duration_ratio_rand, task_duration_ratio_poor)
            .build()
        ),
    )

    actual_pool = pool_params.create_pool_params(markup_config)

    expected_pool = toloka.Pool(
        project_id=project_id,
        private_name=private_name,
        may_contain_adult_content=False,
        will_expire=datetime.datetime(2020, 11, 25),
        reward_per_assignment=reward_per_assignment,
        assignment_max_duration_seconds=int(assignment_duration.total_seconds()) * 5,
        auto_accept_solutions=False,
        auto_accept_period_day=auto_accept_period_day,
        assignments_issuing_config=toloka.Pool.AssignmentsIssuingConfig(issue_task_suites_in_creation_order=False),
        priority=priority,
        defaults=toloka.Pool.Defaults(default_overlap_for_new_tasks=1, default_overlap_for_new_task_suites=1),
    )

    expected_pool.set_mixer_config(real_tasks_count=real_tasks_count, golden_tasks_count=0, training_tasks_count=0)

    expected_pool.set_filter(
        toloka.filter.Languages.in_('RU', verified=True)
        & (toloka.filter.DateOfBirth < int(datetime.datetime(2010, 11, 5).timestamp()))
        & (toloka.filter.DateOfBirth > int(datetime.datetime(1990, 11, 5).timestamp()))
    )

    expected_pool.set_training_requirement(training_requirement=training_requirement)

    expected_pool.quality_control.add_action(
        collector=toloka.collectors.SkippedInRowAssignments(),
        conditions=[toloka.conditions.SkippedInRowCount >= 5],
        action=toloka.actions.RestrictionV2(
            private_comment='Skipped assignments',
            duration=8,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_rand * markup_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, <= 0.2',
            duration=1 * 60 * 24,  # 1 day
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_poor * markup_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, 0.2 < time <= 0.5',
            duration=8 * 60,  # 8 hours
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AnswerCount(),
        conditions=[
            toloka.conditions.AssignmentsAcceptedCount >= int(datetime.timedelta(hours=1) / assignment_duration)
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Completed many assignments, vacation',
            duration=5,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.set_speed_quality_balance(toloka.pool.speed_quality_balance_config.TopPercentageByQuality(percent=90))
    assert actual_pool == expected_pool


@patch.object(human, 'datetime', Mock(wraps=datetime))
@patch.object(pool_params, 'datetime', Mock(wraps=datetime))
def test_check_pool_builder():
    human.datetime.datetime.today.return_value = datetime.datetime(2020, 11, 5)
    pool_params.datetime.datetime.utcnow.return_value = datetime.datetime(2020, 11, 5)
    project_id = '0003'
    private_name = 'custom check config'
    reward_per_assignment = 0.01
    real_tasks_count = 15
    task_duration_hint = datetime.timedelta(seconds=3)
    auto_accept_period_day = 2
    control_tasks_count = 4
    assignment_duration = (control_tasks_count + real_tasks_count) * task_duration_hint
    task_duration_ratio_rand, task_duration_ratio_poor = 0.3, 0.6
    training_requirement = toloka.pool.QualityControl.TrainingRequirement(
        training_pool_id='training pool - check', training_passing_skill_value=65
    )
    priority = 10
    overlap = 5
    check_config = pool_params.ClassificationConfig(
        project_id=project_id,
        private_name=private_name,
        reward_per_assignment=reward_per_assignment,
        task_duration_hint=task_duration_hint,
        real_tasks_count=real_tasks_count,
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN'), worker.LanguageRequirement(lang='DE', verified=True)},
                    age_range=(None, 80),
                    regions={worker.RegionCodes.ITALY, worker.RegionCodes.NETHERLANDS},
                    client_types={toloka.filter.ClientType.ClientType.TOLOKA_APP},
                )
            ],
            training_score=65,
        ),
        training_requirement=training_requirement,
        priority=priority,
        ttl_days=10,
        auto_accept_period_day=auto_accept_period_day,
        may_contain_adult_content=True,
        control_tasks_count=control_tasks_count,
        overlap=overlap,
        control_params=control.Control(
            rules=control.RuleBuilder()
            .add_static_reward(0.5)
            .add_speed_control(task_duration_ratio_rand, task_duration_ratio_poor)
            .add_control_task_control(control_tasks_count, 2, 3)
            .build()
        ),
        work_duration_before_vacation=datetime.timedelta(hours=10),
    )

    actual_pool = pool_params.create_pool_params(check_config)

    expected_pool = toloka.Pool(
        project_id=project_id,
        private_name=private_name,
        may_contain_adult_content=True,
        will_expire=datetime.datetime(2020, 11, 15),
        reward_per_assignment=reward_per_assignment,
        assignment_max_duration_seconds=int(assignment_duration.total_seconds()) * 5,
        auto_accept_solutions=False,
        auto_accept_period_day=auto_accept_period_day,
        assignments_issuing_config=toloka.Pool.AssignmentsIssuingConfig(issue_task_suites_in_creation_order=False),
        priority=priority,
        defaults=toloka.Pool.Defaults(
            default_overlap_for_new_tasks=overlap, default_overlap_for_new_task_suites=overlap
        ),
    )

    expected_pool.set_mixer_config(
        real_tasks_count=real_tasks_count, golden_tasks_count=control_tasks_count, training_tasks_count=0
    )

    expected_pool.set_filter(
        toloka.filter.Languages.in_('DE', verified=True)
        & toloka.filter.Languages.in_('EN')
        & (toloka.filter.RegionByPhone.in_(118) | toloka.filter.RegionByPhone.in_(205))
        & (toloka.filter.DateOfBirth > int(datetime.datetime(1940, 11, 5).timestamp()))
        & (toloka.filter.ClientType == toloka.filter.ClientType.ClientType.TOLOKA_APP)
    )

    expected_pool.set_training_requirement(training_requirement=training_requirement)

    expected_pool.quality_control.add_action(
        collector=toloka.collectors.SkippedInRowAssignments(),
        conditions=[toloka.conditions.SkippedInRowCount >= 5],
        action=toloka.actions.RestrictionV2(
            private_comment='Skipped assignments',
            duration=8,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_rand * check_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, <= 0.3',
            duration=1 * 60 * 24,  # 1 day
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AssignmentSubmitTime(
            fast_submit_threshold_seconds=int(
                task_duration_hint.total_seconds() * task_duration_ratio_poor * check_config.get_tasks_count()
            ),
            history_size=10,
        ),
        conditions=[
            toloka.conditions.TotalSubmittedCount >= 1,
            toloka.conditions.FastSubmittedCount >= 1,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Fast submits, 0.3 < time <= 0.6',
            duration=8 * 60,  # 8 hours
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )

    expected_pool.quality_control.add_action(
        collector=toloka.collectors.GoldenSet(history_size=control_tasks_count),
        conditions=[
            toloka.conditions.TotalAnswersCount >= control_tasks_count,
            toloka.conditions.CorrectAnswersRate < 50.0,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Control tasks: [0, 2) done correctly',
            duration=8 * 60,
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.GoldenSet(history_size=control_tasks_count),
        conditions=[
            toloka.conditions.TotalAnswersCount >= control_tasks_count,
            toloka.conditions.CorrectAnswersRate >= 50.0,
            toloka.conditions.CorrectAnswersRate < 75.0,
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Control tasks: [2, 3) done correctly',
            duration=1 * 60,
            duration_unit=toloka.user_restriction.DurationUnit.MINUTES,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.quality_control.add_action(
        collector=toloka.collectors.AnswerCount(),
        conditions=[
            toloka.conditions.AssignmentsAcceptedCount >= int(datetime.timedelta(hours=10) / assignment_duration)
        ],
        action=toloka.actions.RestrictionV2(
            private_comment='Completed many assignments, vacation',
            duration=5,
            duration_unit=toloka.user_restriction.DurationUnit.HOURS,
            scope=toloka.UserRestriction.PROJECT,
        ),
    )
    expected_pool.set_speed_quality_balance(toloka.pool.speed_quality_balance_config.TopPercentageByQuality(percent=90))
    assert actual_pool == expected_pool


@patch.object(pool_params, 'datetime', Mock(wraps=datetime))
def test_expert_pool_builder():
    pool_params.datetime.datetime.today.return_value = datetime.datetime(2020, 11, 5)
    pool_params.datetime.datetime.utcnow.return_value = datetime.datetime(2020, 11, 5)
    project_id = '0004'
    private_name = 'custom expert config'
    public_description = 'test launch'
    reward_per_assignment = 0.01
    real_tasks_count = 15
    task_duration_hint = datetime.timedelta(seconds=10)
    assignment_duration = real_tasks_count * task_duration_hint
    priority = 10
    overlap = 1
    expert_config = pool_params.ExpertConfig(
        project_id=project_id,
        private_name=private_name,
        public_description=public_description,
        reward_per_assignment=reward_per_assignment,
        task_duration_hint=task_duration_hint,
        real_tasks_count=real_tasks_count,
        priority=priority,
        ttl_days=10,
        worker_filter=worker.ExpertFilter(
            skills=[toloka.Skill(id='experts'), toloka.Skill(id='experts-audio'), toloka.Skill(id='experts-image')]
        ),
    )

    actual_pool = pool_params.create_expert_pool_params(expert_config)

    expected_pool = toloka.Pool(
        project_id=project_id,
        private_name=private_name,
        public_description=public_description,
        may_contain_adult_content=True,
        will_expire=datetime.datetime(2020, 11, 15),
        reward_per_assignment=reward_per_assignment,
        assignment_max_duration_seconds=int(assignment_duration.total_seconds()) * 5,
        auto_accept_solutions=True,
        auto_accept_period_day=14,
        assignments_issuing_config=toloka.Pool.AssignmentsIssuingConfig(issue_task_suites_in_creation_order=False),
        priority=priority,
        defaults=toloka.Pool.Defaults(
            default_overlap_for_new_tasks=overlap, default_overlap_for_new_task_suites=overlap
        ),
    )

    expected_pool.set_mixer_config(real_tasks_count=real_tasks_count, golden_tasks_count=0, training_tasks_count=0)

    expected_pool.set_filter(
        (toloka.filter.Skill('experts') > 0)
        | (toloka.filter.Skill('experts-audio') > 0)
        | (toloka.filter.Skill('experts-image') > 0)
    )
    assert actual_pool == expected_pool



import datetime
from mock import patch, Mock
import pytest
from typing import List

import toloka.client as toloka

from crowdom import mapping, worker
from crowdom.worker import human
from crowdom.objects import Image
from . import lib


class TestModelWorkspace:
    def test_get_solutions(self):
        def func(pool_input_objects: List[mapping.Objects]) -> List[mapping.Objects]:
            results = []
            for (image,) in pool_input_objects:
                index = int(image.url[-5])
                cls = lib.ImageClass.CAT if index % 2 == 0 else lib.ImageClass.DOG
                results.append((cls,))
            return results

        model_ws = worker.ModelWorkspace(
            model=worker.Model(name='my-model', func=func),
            task_mapping=lib.image_classification_mapping,
        )

        assert model_ws.get_solutions(
            [
                (Image(url='https://0.jpg'),),
                (Image(url='https://1.jpg'),),
            ]
        ) == toloka.Assignment(
            tasks=[
                toloka.Task(input_values={'image_url': 'https://0.jpg', 'id': 'Image(url=\'https://0.jpg\')'}),
                toloka.Task(input_values={'image_url': 'https://1.jpg', 'id': 'Image(url=\'https://1.jpg\')'}),
            ],
            solutions=[
                toloka.solution.Solution(output_values={'choice': 'cat'}),
                toloka.solution.Solution(output_values={'choice': 'dog'}),
            ],
            user_id='my-model',
            status=toloka.Assignment.SUBMITTED,
        )

        assert model_ws.solutions_cache == {
            mapping.TaskID((Image(url='https://0.jpg'),)): (lib.ImageClass.CAT,),
            mapping.TaskID((Image(url='https://1.jpg'),)): (lib.ImageClass.DOG,),
        }

        # new task added somewhere in between of previous list, check caching and correctness of solutions order

        assert model_ws.get_solutions(
            [
                (Image(url='https://0.jpg'),),
                (Image(url='https://2.jpg'),),
                (Image(url='https://1.jpg'),),
            ]
        ) == toloka.Assignment(
            tasks=[
                toloka.Task(input_values={'image_url': 'https://0.jpg', 'id': 'Image(url=\'https://0.jpg\')'}),
                toloka.Task(input_values={'image_url': 'https://2.jpg', 'id': 'Image(url=\'https://2.jpg\')'}),
                toloka.Task(input_values={'image_url': 'https://1.jpg', 'id': 'Image(url=\'https://1.jpg\')'}),
            ],
            solutions=[
                toloka.solution.Solution(output_values={'choice': 'cat'}),
                toloka.solution.Solution(output_values={'choice': 'cat'}),
                toloka.solution.Solution(output_values={'choice': 'dog'}),
            ],
            user_id='my-model',
            status=toloka.Assignment.SUBMITTED,
        )

        assert model_ws.solutions_cache == {
            mapping.TaskID((Image(url='https://0.jpg'),)): (lib.ImageClass.CAT,),
            mapping.TaskID((Image(url='https://1.jpg'),)): (lib.ImageClass.DOG,),
            mapping.TaskID((Image(url='https://2.jpg'),)): (lib.ImageClass.CAT,),
        }

    def test_invalid_func(self):
        def func1(pool_input_objects: List[mapping.Objects]):
            return [(lib.ImageClass.CAT,)] * (len(pool_input_objects) + 1)

        def func2(pool_input_objects: List[mapping.Objects]):
            return [lib.ImageClass.CAT] * len(pool_input_objects)

        def func3(pool_input_objects: List[mapping.Objects]):
            return [(lib.ImageClass.CAT, lib.ImageClass.DOG)] * len(pool_input_objects)

        def func4(pool_input_objects: List[mapping.Objects]):
            return [(lib.AudioClass.SPEECH,)] * len(pool_input_objects)

        for func, err_msg in (
            (func1, 'invalid model func: output list length is not equal to input'),
            (func2, 'invalid model func: output objects are not tuple'),
            (func3, 'invalid model func: output objects count do not match task function'),
            (
                func4,
                'invalid model func: object type mismatch, expected <enum \'ImageClass\'>, actual <enum \'AudioClass\'>',
            ),
        ):
            model_ws = worker.ModelWorkspace(
                model=worker.Model(name='my-model', func=func),
                task_mapping=lib.image_classification_mapping,
            )
            with pytest.raises(AssertionError) as e:
                model_ws.get_solutions(
                    [
                        (Image(url='https://0.jpg'),),
                        (Image(url='https://1.jpg'),),
                    ]
                )
            assert str(e.value) == err_msg


class TestHumanFilters:
    def test_expert_filter(self):
        assert worker.ExpertFilter(
            skills=[toloka.Skill(id='expert'), toloka.Skill(id='expert_my-task_EN')]
        ).to_toloka_filter() == (toloka.filter.Skill('expert') > 0) | (toloka.filter.Skill('expert_my-task_EN') > 0)

    @patch.object(human, 'datetime', Mock(wraps=datetime))
    def test_worker_filter(self):
        human.datetime.datetime.today.return_value = datetime.datetime(2020, 11, 5)

        assert worker.WorkerFilter(
            training_score=50,
            filters=[
                worker.WorkerFilter.Params(
                    langs=set(),
                    age_range=(None, 80),
                    client_types={toloka.filter.ClientType.ClientType.BROWSER},
                ),
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.TURKEY, worker.RegionCodes.SPAIN},
                    age_range=(24, 40),
                ),
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN'), worker.LanguageRequirement(lang='DE', verified=True)},
                    regions={worker.RegionCodes.ITALY},
                    client_types={
                        toloka.filter.ClientType.ClientType.TOLOKA_APP,
                        toloka.filter.ClientType.ClientType.BROWSER,
                    },
                ),
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='DE')},
                    age_range=(24, None),
                ),
            ],
        ).to_toloka_filter() == (
            (toloka.filter.DateOfBirth > int(datetime.datetime(1940, 11, 5).timestamp()))
            & (toloka.filter.ClientType == toloka.filter.ClientType.ClientType.BROWSER)
        ) | (
            toloka.filter.Languages.in_('EN', verified=True)
            & (toloka.filter.RegionByPhone.in_(204) | toloka.filter.RegionByPhone.in_(983))
            & (toloka.filter.DateOfBirth < int(datetime.datetime(1996, 11, 5).timestamp()))
            & (toloka.filter.DateOfBirth > int(datetime.datetime(1980, 11, 5).timestamp()))
        ) | (
            toloka.filter.Languages.in_('DE', verified=True)
            & toloka.filter.Languages.in_('EN')
            & toloka.filter.RegionByPhone.in_(205)
            & (toloka.filter.DateOfBirth < int(datetime.datetime(2002, 11, 5).timestamp()))
            & (
                (toloka.filter.ClientType == toloka.filter.ClientType.ClientType.BROWSER)
                | (toloka.filter.ClientType == toloka.filter.ClientType.ClientType.TOLOKA_APP)
            )
        ) | (
            toloka.filter.Languages.in_('DE')
            & (toloka.filter.DateOfBirth < int(datetime.datetime(1996, 11, 5).timestamp()))
        )

        assert worker.WorkerFilter(training_score=None, filters=[]).to_toloka_filter() is None

    def test_invalid_training_score(self):
        with pytest.raises(AssertionError):
            worker.WorkerFilter(training_score=101, filters=[])



from typing import Dict

import toloka.client.project.template_builder as tb
import toloka.client as toloka

from crowdom import base, project as project_params, objects, task_spec as spec
from crowdom.mapping import TaskMapping, ObjectMapping

from . import lib


class Animal(base.Class):
    DOG = 'dog'
    CAT = 'cat'

    @classmethod
    def labels(cls) -> Dict[base.Class, Dict[str, str]]:
        return {value: {'EN': str(value.name)} for value in cls.possible_instances()}


class ManyOptions(base.Class):
    O1 = '1'
    O2 = '2'
    O3 = '3'
    O4 = '4'
    O5 = '5'
    O6 = '6'

    @classmethod
    def labels(cls) -> Dict[base.Class, Dict[str, str]]:
        return {value: {'EN': f'option {value.value}'} for value in cls.possible_instances()}


def test_label_view():
    for obj_meta, is_output, expected_view in [
        (
            base.ClassMeta(name='cls', type=Animal),
            True,
            tb.RadioGroupFieldV1(
                data=tb.OutputData(path='cls'),
                options=[
                    tb.GroupFieldOption(value='dog', label='DOG'),
                    tb.GroupFieldOption(value='cat', label='CAT'),
                ],
                disabled=False,
            ),
        ),
        (
            base.ClassMeta(name='cls', type=Animal),
            False,
            tb.RadioGroupFieldV1(
                data=tb.OutputData(path='cls'),
                options=[
                    tb.GroupFieldOption(value='dog', label='DOG'),
                    tb.GroupFieldOption(value='cat', label='CAT'),
                ],
                disabled=True,
            ),
        ),
        (
            base.ClassMeta(name='cls', type=ManyOptions),
            True,
            tb.SelectFieldV1(
                data=tb.OutputData(path='cls'),
                options=[
                    tb.SelectFieldV1.Option(label='option 1', value='1'),
                    tb.SelectFieldV1.Option(label='option 2', value='2'),
                    tb.SelectFieldV1.Option(label='option 3', value='3'),
                    tb.SelectFieldV1.Option(label='option 4', value='4'),
                    tb.SelectFieldV1.Option(label='option 5', value='5'),
                    tb.SelectFieldV1.Option(label='option 6', value='6'),
                ],
            ),
        ),
        (
            base.ClassMeta(name='cls', type=ManyOptions),
            False,
            tb.IfHelperV1(
                condition=tb.EqualsConditionV1(
                    data=tb.InputData(path='cls'),
                    to='1',
                ),
                then=tb.TextViewV1(content='option 1'),
                else_=tb.IfHelperV1(
                    condition=tb.EqualsConditionV1(
                        data=tb.InputData(path='cls'),
                        to='2',
                    ),
                    then=tb.TextViewV1(content='option 2'),
                    else_=tb.IfHelperV1(
                        condition=tb.EqualsConditionV1(
                            data=tb.InputData(path='cls'),
                            to='3',
                        ),
                        then=tb.TextViewV1(content='option 3'),
                        else_=tb.IfHelperV1(
                            condition=tb.EqualsConditionV1(
                                data=tb.InputData(path='cls'),
                                to='4',
                            ),
                            then=tb.TextViewV1(content='option 4'),
                            else_=tb.IfHelperV1(
                                condition=tb.EqualsConditionV1(
                                    data=tb.InputData(path='cls'),
                                    to='5',
                                ),
                                then=tb.TextViewV1(content='option 5'),
                                else_=tb.IfHelperV1(
                                    condition=tb.EqualsConditionV1(
                                        data=tb.InputData(path='cls'),
                                        to='6',
                                    ),
                                    then=tb.TextViewV1(content='option 6'),
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        ),
        (
            base.ClassMeta(
                name='cls',
                type=Animal,
                input_display_type=base.LabelsDisplayType.MONO,
                text_format=base.TextFormat.MARKDOWN,
            ),
            False,
            tb.IfHelperV1(
                condition=tb.EqualsConditionV1(
                    data=tb.InputData(path='cls'),
                    to='dog',
                ),
                then=tb.MarkdownViewV1(content='DOG'),
                else_=tb.IfHelperV1(
                    condition=tb.EqualsConditionV1(
                        data=tb.InputData(path='cls'),
                        to='cat',
                    ),
                    then=tb.MarkdownViewV1(content='CAT'),
                ),
            ),
        ),
        (
            base.ClassMeta(
                name='cls', type=Animal, title=base.Title(text=base.LocalizedString({'EN': 'choose animal'}))
            ),
            True,
            tb.GroupViewV1(
                content=tb.ListViewV1(
                    items=[
                        tb.TextViewV1(content='choose animal'),
                        tb.RadioGroupFieldV1(
                            data=tb.OutputData(path='cls'),
                            options=[
                                tb.GroupFieldOption(value='dog', label='DOG'),
                                tb.GroupFieldOption(value='cat', label='CAT'),
                            ],
                            disabled=False,
                        ),
                    ],
                ),
            ),
        ),
        (
            base.ClassMeta(
                type=ManyOptions,
                name='cls',
                available_labels=base.create_available_labels_if(
                    name='question',
                    value_labels=[
                        ('q1', [ManyOptions.O6]),
                        ('q2', [ManyOptions.O1, ManyOptions.O5]),
                        ('q3', [ManyOptions.O2, ManyOptions.O3]),
                    ],
                ),
            ),
            True,
            tb.IfHelperV1(
                condition=tb.EqualsConditionV1(
                    data=tb.InputData(path='question_f'),
                    to='q1',
                ),
                then=tb.RadioGroupFieldV1(
                    data=tb.OutputData(path='cls'),
                    options=[
                        tb.GroupFieldOption(value='6', label='option 6'),
                    ],
                    disabled=False,
                ),
                else_=tb.IfHelperV1(
                    condition=tb.EqualsConditionV1(
                        data=tb.InputData(path='question_f'),
                        to='q2',
                    ),
                    then=tb.RadioGroupFieldV1(
                        data=tb.OutputData(path='cls'),
                        options=[
                            tb.GroupFieldOption(value='1', label='option 1'),
                            tb.GroupFieldOption(value='5', label='option 5'),
                        ],
                        disabled=False,
                    ),
                    else_=tb.IfHelperV1(
                        condition=tb.EqualsConditionV1(
                            data=tb.InputData(path='question_f'),
                            to='q3',
                        ),
                        then=tb.RadioGroupFieldV1(
                            data=tb.OutputData(path='cls'),
                            options=[
                                tb.GroupFieldOption(value='2', label='option 2'),
                                tb.GroupFieldOption(value='3', label='option 3'),
                            ],
                            disabled=False,
                        ),
                    ),
                ),
            ),
        ),
        (
            base.ClassMeta(
                type=Animal,
                name='cls',
                available_labels=base.create_available_labels_if(
                    name='question',
                    value_labels=[
                        ('q1', [Animal.CAT]),
                        ('q2', [Animal.CAT, Animal.DOG]),
                    ],
                ),
                input_display_type=base.LabelsDisplayType.MONO,
                title=base.Title(text=base.LocalizedString({'EN': 'answer question'})),
            ),
            False,
            tb.IfHelperV1(
                condition=tb.EqualsConditionV1(
                    to='q1',
                    data=tb.InputData(path='question_f'),
                ),
                then=tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='answer question'),
                            tb.IfHelperV1(
                                condition=tb.EqualsConditionV1(
                                    to='dog',
                                    data=tb.InputData(path='cls'),
                                ),
                                then=tb.TextViewV1(content='DOG'),
                                else_=tb.IfHelperV1(
                                    condition=tb.EqualsConditionV1(
                                        to='cat',
                                        data=tb.InputData(path='cls'),
                                    ),
                                    then=tb.TextViewV1(content='CAT'),
                                ),
                            ),
                        ],
                    ),
                ),
                else_=tb.IfHelperV1(
                    condition=tb.EqualsConditionV1(
                        to='q2',
                        data=tb.InputData(path='question_f'),
                    ),
                    then=tb.GroupViewV1(
                        content=tb.ListViewV1(
                            items=[
                                tb.TextViewV1(content='answer question'),
                                tb.IfHelperV1(
                                    condition=tb.EqualsConditionV1(
                                        to='dog',
                                        data=tb.InputData(path='cls'),
                                    ),
                                    then=tb.TextViewV1(content='DOG'),
                                    else_=tb.IfHelperV1(
                                        condition=tb.EqualsConditionV1(
                                            to='cat',
                                            data=tb.InputData(path='cls'),
                                        ),
                                        then=tb.TextViewV1(content='CAT'),
                                    ),
                                ),
                            ],
                        ),
                    ),
                ),
            ),
        ),
    ]:
        actual_view = project_params.gen_label_view(
            data=tb.OutputData(path='cls'),
            is_output=is_output,
            obj_meta=obj_meta,
            condition_context=base.ConditionContext(
                name_to_arg_data={
                    'cls': base.ConditionContext.ArgData(final_name='cls', is_output=is_output),
                    'question': base.ConditionContext.ArgData(final_name='question_f', is_output=False),
                }
            ),
            lang='EN',
        )
        assert actual_view == expected_view


def test_validation_view():
    for validation, placeholder, expected_view in [
        (
            None,
            None,
            tb.TextareaFieldV1(
                data=tb.OutputData(path='text'),
            ),
        ),
        (
            objects.TextValidation(regex={'EN': '^[a-z]+$'}, hint={'EN': 'One lowercase word only'}),
            base.LocalizedString({'EN': 'Placeholder text'}),
            tb.TextareaFieldV1(
                data=tb.OutputData(path='text'),
                placeholder='Placeholder text',
                validation=tb.SchemaConditionV1(
                    schema={'type': 'string', 'pattern': '^[a-z]+$'},
                    hint='One lowercase word only',
                ),
            ),
        ),
    ]:
        actual_view = project_params.gen_text_output_view(
            data=tb.OutputData(path='text'),
            validation=validation,
            placeholder=placeholder,
            lang='EN',
        )
        assert actual_view == expected_view


def test_obj_type_indexes():
    types = (objects.Text, objects.Audio, objects.Image, objects.Text, objects.Image, objects.Text)
    expected_indexes = ['_1', '', '_1', '_2', '_2', '_3']
    assert project_params.get_obj_type_indexes(types) == expected_indexes


# Common cases we need to cover
# [x] Multiple input/output objects
# [x] Localization fallback to EN
# [x] Label as input
# [x] Multiple objects of same type in input/output
# [x] Expert annotation of task
# [x] Expert annotation of solution
# [x] Overlapping types in input and output
# [x] Definition through ObjectMeta
# [x] Titles for Label types
# [ ] Titles for Text type
# [x] Custom Text format
# [x] Text regex validation
# [ ] Display input Class as text
# [x] Conditional options for Class
# [ ] Title along with conditional options
# [x] Condition on input argument
# [ ] Condition on output argument
# [x] Condition on argument moved from output to input
# [x] Conditions transform for SbS function
# [ ] Annotation function expert labeling types
# [ ] Cover all objects types as input and output


def test_sbs_project_compare_layout_multiple_objects():
    # [x] Multiple input/output objects
    # [x] Multiple objects of same type in input/output

    builder = project_params.SbSBuilder(base.SbSFunction(inputs=(objects.Audio, objects.Text, objects.Text)), lang='EN')
    sbs_view, sbs_mapping = builder.get_default_scenario()
    assert sbs_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Audio, name='audio'), (('url', 'audio_a'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_1'), (('text', 'text_1_a'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_2'), (('text', 'text_2_a'),)),
            ObjectMapping(base.ObjectMeta(objects.Audio, name='audio'), (('url', 'audio_b'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_1'), (('text', 'text_1_b'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_2'), (('text', 'text_2_b'),)),
        ),
        output_mapping=(ObjectMapping(base.ObjectMeta(base.SbSChoice, name='choice'), (('value', 'choice'),)),),
    )
    assert sbs_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.CompareLayoutV1(
                    items=[
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.AudioViewV1(
                                        url=tb.InputData(path='audio_a'), validation=tb.PlayedFullyConditionV1()
                                    ),
                                    tb.TextViewV1(content=tb.InputData(path='text_1_a')),
                                    tb.TextViewV1(content=tb.InputData(path='text_2_a')),
                                ]
                            )
                        ),
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.AudioViewV1(
                                        url=tb.InputData(path='audio_b'), validation=tb.PlayedFullyConditionV1()
                                    ),
                                    tb.TextViewV1(content=tb.InputData(path='text_1_b')),
                                    tb.TextViewV1(content=tb.InputData(path='text_2_b')),
                                ]
                            )
                        ),
                    ],
                    common_controls=tb.ColumnsLayoutV1(
                        items=[
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='choice'),
                                disabled=False,
                                options=[tb.GroupFieldOption(label='Left option', value='a')],
                            ),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='choice'),
                                disabled=False,
                                options=[tb.GroupFieldOption(label='Right option', value='b')],
                            ),
                        ],
                        min_width=10.0,
                        vertical_align='top',
                        validation=tb.RequiredConditionV1(
                            hint='Choose one of the options', data=tb.OutputData(path='choice')
                        ),
                    ),
                    wide_common_controls=True,
                )
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice'), payload='a')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice'), payload='b')),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )


def test_sbs_project_sbs_layout_with_hint():
    builder = project_params.SbSBuilder(
        base.SbSFunction(inputs=(objects.Video,), hints=(objects.Image, objects.Image)), lang='RU'
    )
    sbs_view, sbs_mapping = builder.get_default_scenario()
    assert sbs_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image_1'), (('url', 'image_1_hint'),)),
            ObjectMapping(base.ObjectMeta(objects.Image, name='image_2'), (('url', 'image_2_hint'),)),
            ObjectMapping(base.ObjectMeta(objects.Video, name='video'), (('url', 'video_a'),)),
            ObjectMapping(base.ObjectMeta(objects.Video, name='video'), (('url', 'video_b'),)),
        ),
        output_mapping=(ObjectMapping(base.ObjectMeta(base.SbSChoice, name='choice'), (('value', 'choice'),)),),
    )
    assert sbs_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.ImageViewV1(url=tb.InputData(path='image_1_hint')),
                tb.ImageViewV1(url=tb.InputData(path='image_2_hint')),
                tb.SideBySideLayoutV1(
                    items=[
                        tb.VideoViewV1(url=tb.InputData(path='video_a')),
                        tb.VideoViewV1(url=tb.InputData(path='video_b')),
                    ],
                    controls=tb.RadioGroupFieldV1(
                        data=tb.OutputData(path='choice'),
                        disabled=False,
                        options=[
                            tb.GroupFieldOption(label='Левый вариант', value='a'),
                            tb.GroupFieldOption(label='Правый вариант', value='b'),
                        ],
                        validation=tb.RequiredConditionV1(
                            hint='Выберите один из вариантов', data=tb.OutputData(path='choice')
                        ),
                    ),
                ),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice'), payload='a')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice'), payload='b')),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=1200.0),
                notifications=[
                    tb.TextViewV1(content='Перед выполнением проверьте, что все медиа-элементы загрузились.'),
                    tb.TextViewV1(content='Если нет хотя бы одного медиа-элемента, перезагрузите страницу.'),
                ],
            ),
        ],
    )


def test_sbs_project_expert_labeling_of_solution():
    # [x] Expert annotation of solution
    # [x] Overlapping types in input and output

    builder = project_params.SbSBuilder(
        base.SbSFunction(inputs=(objects.Audio, Animal), hints=(objects.Text,)), lang='EN'
    )

    sbs_view, sbs_mapping = builder.get_expert_labeling_of_solved_tasks()
    assert sbs_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text_hint'),)),
            ObjectMapping(base.ObjectMeta(objects.Audio, name='audio'), (('url', 'audio_a'),)),
            ObjectMapping(base.ObjectMeta(Animal, name='choice_1'), (('value', 'choice_1_a'),)),
            ObjectMapping(base.ObjectMeta(objects.Audio, name='audio'), (('url', 'audio_b'),)),
            ObjectMapping(base.ObjectMeta(Animal, name='choice_1'), (('value', 'choice_1_b'),)),
            ObjectMapping(base.ObjectMeta(base.SbSChoice, name='choice_2'), (('value', 'choice_2'),)),
        ),
        output_mapping=(
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='_ok',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the task (and the solution, if given) correct?',
                                'RU': 'Корректно ли задание (и решение, при наличии)?',
                            }
                        )
                    ),
                ),
                (('ok', '_ok'),),
            ),
            ObjectMapping(
                base.ObjectMeta(
                    objects.Text,
                    name='_comment',
                    title=base.Title(text=base.LocalizedString(lang_to_text={'EN': 'comment', 'RU': 'комментарий'})),
                    required=False,
                ),
                (('text', '_comment'),),
            ),
        ),
    )
    assert sbs_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text_hint')),
                tb.CompareLayoutV1(
                    items=[
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.AudioViewV1(
                                        url=tb.InputData(path='audio_a'), validation=tb.PlayedFullyConditionV1()
                                    ),
                                    tb.RadioGroupFieldV1(
                                        data=tb.InternalData(
                                            path='choice_1_a', default=tb.InputData(path='choice_1_a')
                                        ),
                                        disabled=True,
                                        options=[
                                            tb.GroupFieldOption(value='dog', label='DOG'),
                                            tb.GroupFieldOption(value='cat', label='CAT'),
                                        ],
                                    ),
                                ]
                            )
                        ),
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.AudioViewV1(
                                        url=tb.InputData(path='audio_b'), validation=tb.PlayedFullyConditionV1()
                                    ),
                                    tb.RadioGroupFieldV1(
                                        data=tb.InternalData(
                                            path='choice_1_b', default=tb.InputData(path='choice_1_b')
                                        ),
                                        disabled=True,
                                        options=[
                                            tb.GroupFieldOption(value='dog', label='DOG'),
                                            tb.GroupFieldOption(value='cat', label='CAT'),
                                        ],
                                    ),
                                ]
                            )
                        ),
                    ],
                    common_controls=tb.ColumnsLayoutV1(
                        items=[
                            tb.RadioGroupFieldV1(
                                data=tb.InternalData(path='choice_2', default=tb.InputData(path='choice_2')),
                                disabled=True,
                                options=[tb.GroupFieldOption(label='Left option', value='a')],
                            ),
                            tb.RadioGroupFieldV1(
                                data=tb.InternalData(path='choice_2', default=tb.InputData(path='choice_2')),
                                disabled=True,
                                options=[tb.GroupFieldOption(label='Right option', value='b')],
                            ),
                        ],
                        min_width=10.0,
                        vertical_align='top',
                    ),
                    wide_common_controls=True,
                ),
                tb.MarkdownViewV1(content='---'),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Is the task (and the solution, if given) correct?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='_ok'),
                                disabled=False,
                                options=[
                                    tb.GroupFieldOption(value=True, label='Yes'),
                                    tb.GroupFieldOption(value=False, label='No'),
                                ],
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options',
                                    data=tb.OutputData(path='_ok'),
                                ),
                            ),
                        ]
                    )
                ),
                tb.TextareaFieldV1(data=tb.OutputData(path='_comment'), placeholder='comment'),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('_ok'), payload=True)),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('_ok'), payload=False)),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )


def test_classification_project():
    # [x] Localization fallback to EN
    # [x] Multiple objects of same type in input/output
    # [x] Label as input
    # [x] Overlapping types in input and output

    builder = project_params.ClassificationBuilder(
        base.ClassificationFunction(inputs=(objects.Text, objects.Text, Animal), cls=Animal), lang='RU'
    )

    classification_view, classification_mapping = builder.get_default_scenario()

    assert classification_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_1'), (('text', 'text_1'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_2'), (('text', 'text_2'),)),
            ObjectMapping(base.ObjectMeta(Animal, name='choice_1'), (('value', 'choice_1'),)),
        ),
        output_mapping=(ObjectMapping(base.ObjectMeta(Animal, name='choice_2'), (('value', 'choice_2'),)),),
    )
    assert classification_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text_1')),
                tb.TextViewV1(content=tb.InputData(path='text_2')),
                tb.RadioGroupFieldV1(
                    data=tb.InternalData(path='choice_1', default=tb.InputData(path='choice_1')),
                    options=[
                        tb.GroupFieldOption(value='dog', label='DOG'),  # fallback to EN
                        tb.GroupFieldOption(value='cat', label='CAT'),
                    ],  # fallback to EN
                    disabled=True,
                ),
                tb.RadioGroupFieldV1(
                    data=tb.OutputData(path='choice_2'),
                    options=[
                        tb.GroupFieldOption(value='dog', label='DOG'),  # fallback to EN
                        tb.GroupFieldOption(value='cat', label='CAT'),
                    ],  # fallback to EN
                    disabled=False,
                    validation=tb.RequiredConditionV1(
                        hint='Выберите один из вариантов', data=tb.OutputData(path='choice_2')
                    ),
                ),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice_2'), payload='dog')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice_2'), payload='cat')),
            tb.TolokaPluginV1(layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0)),
        ],
    )


def test_classification_project_expert_labeling_of_tasks():
    # [x] Expert annotation of task
    builder = project_params.ClassificationBuilder(
        base.ClassificationFunction(inputs=(objects.Text,), cls=Animal), lang='EN'
    )

    classification_view, classification_mapping = builder.get_expert_labeling_of_tasks()
    assert classification_mapping == TaskMapping(
        input_mapping=(ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text'),)),),
        output_mapping=(
            ObjectMapping(base.ObjectMeta(Animal, name='choice'), (('value', 'choice'),)),
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='_ok',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the task (and the solution, if given) correct?',
                                'RU': 'Корректно ли задание (и решение, при наличии)?',
                            }
                        )
                    ),
                ),
                (('ok', '_ok'),),
            ),
            ObjectMapping(
                base.ObjectMeta(
                    objects.Text,
                    name='_comment',
                    title=base.Title(text=base.LocalizedString(lang_to_text={'EN': 'comment', 'RU': 'комментарий'})),
                    required=False,
                ),
                (('text', '_comment'),),
            ),
        ),
    )
    assert classification_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text')),
                tb.RadioGroupFieldV1(
                    data=tb.OutputData(path='choice'),
                    options=[
                        tb.GroupFieldOption(value='dog', label='DOG'),
                        tb.GroupFieldOption(value='cat', label='CAT'),
                    ],
                    disabled=False,
                    validation=tb.RequiredConditionV1(
                        hint='Choose one of the options', data=tb.OutputData(path='choice')
                    ),
                ),
                tb.MarkdownViewV1('---'),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Is the task (and the solution, if given) correct?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='_ok'),
                                disabled=False,
                                options=[
                                    tb.GroupFieldOption(value=True, label='Yes'),
                                    tb.GroupFieldOption(value=False, label='No'),
                                ],
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options',
                                    data=tb.OutputData(path='_ok'),
                                ),
                            ),
                        ]
                    )
                ),
                tb.TextareaFieldV1(data=tb.OutputData(path='_comment'), placeholder='comment'),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice'), payload='dog')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice'), payload='cat')),
            tb.HotkeysPluginV1(key_3=tb.SetActionV1(data=tb.OutputData('_ok'), payload=True)),
            tb.HotkeysPluginV1(key_4=tb.SetActionV1(data=tb.OutputData('_ok'), payload=False)),
            tb.TolokaPluginV1(layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0)),
        ],
    )


def test_annotation_project():
    # [x] Label as input
    # [x] Overlapping types in input and output
    task_spec = spec.AnnotationTaskSpec(
        task_spec=base.TaskSpec(
            id=None,  # noqa
            instruction=None,  # noqa
            name=base.LocalizedString({'EN': 'name'}),
            description=base.LocalizedString({'EN': 'desc'}),
            function=base.AnnotationFunction(
                inputs=(objects.Image, objects.Text), outputs=(objects.Text, lib.ImageClass)
            ),
        ),
        lang='EN',
    )

    annotation_view, annotation_mapping = task_spec.view, task_spec.task_mapping
    check_view, check_mapping = task_spec.check.view, task_spec.check.task_mapping

    assert annotation_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_1'), (('text', 'text_1'),)),
        ),
        output_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_2'), (('text', 'text_2'),)),
            ObjectMapping(base.ObjectMeta(lib.ImageClass, name='choice'), (('value', 'choice'),)),
        ),
    )
    assert check_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_1'), (('text', 'text_1'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text_2'), (('text', 'text_2'),)),
            ObjectMapping(base.ObjectMeta(lib.ImageClass, name='choice'), (('value', 'choice'),)),
        ),
        output_mapping=(
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='eval',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the solution accurate?',
                                'KK': 'Тапсырманың шешімі дұрыс па?',
                                'RU': 'Решение верное?',
                                'TR': 'Görevin kararı doğru mu?',
                            }
                        )
                    ),
                ),
                (('ok', 'eval'),),
            ),
        ),
    )
    assert annotation_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.ImageViewV1(url=tb.InputData(path='image')),
                tb.TextViewV1(content=tb.InputData(path='text_1')),
                tb.TextareaFieldV1(data=tb.OutputData(path='text_2')),
                tb.RadioGroupFieldV1(
                    data=tb.OutputData(path='choice'),
                    disabled=False,
                    options=[
                        tb.GroupFieldOption(value='dog', label='dog'),
                        tb.GroupFieldOption(value='cat', label='cat'),
                        tb.GroupFieldOption(value='crow', label='crow'),
                    ],
                    validation=tb.RequiredConditionV1(
                        hint='Choose one of the options', data=tb.OutputData(path='choice')
                    ),
                ),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice'), payload='dog')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice'), payload='cat')),
            tb.HotkeysPluginV1(key_3=tb.SetActionV1(data=tb.OutputData('choice'), payload='crow')),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )
    assert check_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.ImageViewV1(url=tb.InputData(path='image')),
                tb.TextViewV1(content=tb.InputData(path='text_1')),
                tb.TextViewV1(content=tb.InputData(path='text_2')),
                tb.RadioGroupFieldV1(
                    data=tb.InternalData(path='choice', default=tb.InputData(path='choice')),
                    options=[
                        tb.GroupFieldOption(value='dog', label='dog'),
                        tb.GroupFieldOption(value='cat', label='cat'),
                        tb.GroupFieldOption(value='crow', label='crow'),
                    ],
                    disabled=True,
                ),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Is the solution accurate?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='eval'),
                                options=[
                                    tb.GroupFieldOption(value=True, label='Yes'),
                                    tb.GroupFieldOption(value=False, label='No'),
                                ],
                                disabled=False,
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options', data=tb.OutputData(path='eval')
                                ),
                            ),
                        ]
                    )
                ),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('eval'), payload=True)),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('eval'), payload=False)),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )


def test_classification_project_with_object_metas_and_titles_and_custom_text_format():
    # [x] Definition through ObjectMeta
    # [x] Titles for Label types
    # [x] Custom Text format

    builder = project_params.ClassificationBuilder(
        base.ClassificationFunction(
            inputs=(objects.TextMeta(objects.Text, name='description', format=base.TextFormat.MARKDOWN),),
            cls=base.ClassMeta(
                Animal,
                name='animal',
                title=base.Title(
                    text=base.LocalizedString({'EN': 'Which animal best describes this text?'}),
                    format=base.TextFormat.MARKDOWN,
                ),
            ),
        ),
        lang='EN',
    )

    classification_view, classification_mapping = builder.get_default_scenario()

    assert classification_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(
                objects.TextMeta(objects.Text, name='description', format=base.TextFormat.MARKDOWN),
                (('text', 'description'),),
            ),
        ),
        output_mapping=(
            ObjectMapping(
                base.ClassMeta(
                    Animal,
                    name='animal',
                    title=base.Title(
                        text=base.LocalizedString({'EN': 'Which animal best describes this text?'}),
                        format=base.TextFormat.MARKDOWN,
                    ),
                ),
                (('value', 'animal'),),
            ),
        ),
    )
    assert classification_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.MarkdownViewV1(content=tb.InputData(path='description')),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.MarkdownViewV1(content='Which animal best describes this text?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='animal'),
                                options=[
                                    tb.GroupFieldOption(value='dog', label='DOG'),
                                    tb.GroupFieldOption(value='cat', label='CAT'),
                                ],
                                disabled=False,
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options', data=tb.OutputData(path='animal')
                                ),
                            ),
                        ],
                    ),
                ),
            ],
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('animal'), payload='dog')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('animal'), payload='cat')),
            tb.TolokaPluginV1(layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0)),
        ],
    )


class Topic(base.Class):
    MOVIE = 'movie'
    NEWS = 'news'


class Review(base.Class):
    POSITIVE = 'positive'
    NEGATIVE = 'negative'
    NEUTRAL = 'neutral'
    SKEPTICAL = 'skeptical'
    APPROVING = 'approving'


options_if = base.create_available_labels_if(
    'topic',
    [
        (Topic.MOVIE.value, [Review.POSITIVE, Review.NEGATIVE, Review.NEUTRAL]),
        (Topic.NEWS.value, [Review.SKEPTICAL, Review.APPROVING]),
    ],
)


def test_sbs_project_with_conditional_options():
    # [x] Conditional options for Class
    # [x] Condition on input argument
    # [x] Conditions transform for SbS function

    builder = project_params.SbSBuilder(
        base.SbSFunction(
            inputs=(
                base.ClassMeta(Topic, name='topic'),
                base.ClassMeta(Review, name='review', available_labels=options_if),
            ),
        ),
        lang='RU',
    )
    sbs_view, sbs_mapping = builder.get_default_scenario()
    assert sbs_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ClassMeta(Topic, name='topic'), (('value', 'topic_a'),)),
            ObjectMapping(base.ClassMeta(Review, name='review', available_labels=options_if), (('value', 'review_a'),)),
            ObjectMapping(base.ClassMeta(Topic, name='topic'), (('value', 'topic_b'),)),
            ObjectMapping(base.ClassMeta(Review, name='review', available_labels=options_if), (('value', 'review_b'),)),
        ),
        output_mapping=(ObjectMapping(base.ObjectMeta(base.SbSChoice, name='choice'), (('value', 'choice'),)),),
    )
    assert sbs_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.CompareLayoutV1(
                    items=[
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.RadioGroupFieldV1(
                                        data=tb.InternalData(path='topic_a', default=tb.InputData(path='topic_a')),
                                        options=[
                                            tb.GroupFieldOption(value='movie', label='movie'),
                                            tb.GroupFieldOption(value='news', label='news'),
                                        ],
                                        disabled=True,
                                    ),
                                    tb.IfHelperV1(
                                        condition=tb.EqualsConditionV1(
                                            data=tb.InputData(path='topic_a'),
                                            to='movie',
                                        ),
                                        then=tb.RadioGroupFieldV1(
                                            data=tb.InternalData(
                                                path='review_a', default=tb.InputData(path='review_a')
                                            ),
                                            options=[
                                                tb.GroupFieldOption(value='positive', label='positive'),
                                                tb.GroupFieldOption(value='negative', label='negative'),
                                                tb.GroupFieldOption(value='neutral', label='neutral'),
                                            ],
                                            disabled=True,
                                        ),
                                        else_=tb.IfHelperV1(
                                            condition=tb.EqualsConditionV1(
                                                data=tb.InputData(path='topic_a'),
                                                to='news',
                                            ),
                                            then=tb.RadioGroupFieldV1(
                                                data=tb.InternalData(
                                                    path='review_a', default=tb.InputData(path='review_a')
                                                ),
                                                options=[
                                                    tb.GroupFieldOption(value='skeptical', label='skeptical'),
                                                    tb.GroupFieldOption(value='approving', label='approving'),
                                                ],
                                                disabled=True,
                                            ),
                                        ),
                                    ),
                                ]
                            )
                        ),
                        tb.CompareLayoutItem(
                            content=tb.ListViewV1(
                                items=[
                                    tb.RadioGroupFieldV1(
                                        data=tb.InternalData(path='topic_b', default=tb.InputData(path='topic_b')),
                                        options=[
                                            tb.GroupFieldOption(value='movie', label='movie'),
                                            tb.GroupFieldOption(value='news', label='news'),
                                        ],
                                        disabled=True,
                                    ),
                                    tb.IfHelperV1(
                                        condition=tb.EqualsConditionV1(
                                            data=tb.InputData(path='topic_b'),
                                            to='movie',
                                        ),
                                        then=tb.RadioGroupFieldV1(
                                            data=tb.InternalData(
                                                path='review_b', default=tb.InputData(path='review_b')
                                            ),
                                            options=[
                                                tb.GroupFieldOption(value='positive', label='positive'),
                                                tb.GroupFieldOption(value='negative', label='negative'),
                                                tb.GroupFieldOption(value='neutral', label='neutral'),
                                            ],
                                            disabled=True,
                                        ),
                                        else_=tb.IfHelperV1(
                                            condition=tb.EqualsConditionV1(
                                                data=tb.InputData(path='topic_b'),
                                                to='news',
                                            ),
                                            then=tb.RadioGroupFieldV1(
                                                data=tb.InternalData(
                                                    path='review_b', default=tb.InputData(path='review_b')
                                                ),
                                                options=[
                                                    tb.GroupFieldOption(value='skeptical', label='skeptical'),
                                                    tb.GroupFieldOption(value='approving', label='approving'),
                                                ],
                                                disabled=True,
                                            ),
                                        ),
                                    ),
                                ]
                            )
                        ),
                    ],
                    common_controls=tb.ColumnsLayoutV1(
                        items=[
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='choice'),
                                disabled=False,
                                options=[tb.GroupFieldOption(label='Левый вариант', value='a')],
                            ),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='choice'),
                                disabled=False,
                                options=[tb.GroupFieldOption(label='Правый вариант', value='b')],
                            ),
                        ],
                        min_width=10.0,
                        vertical_align='top',
                        validation=tb.RequiredConditionV1(
                            hint='Выберите один из вариантов', data=tb.OutputData(path='choice')
                        ),
                    ),
                    wide_common_controls=True,
                )
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('choice'), payload='a')),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('choice'), payload='b')),
            tb.TolokaPluginV1(layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0)),
        ],
    )


def test_annotation_project_with_conditional_options_solved_tasks_verification():
    # [x] Condition on argument moved from output to input
    # [x] Conditional options for Class

    task_spec = base.TaskSpec(
        id=None,  # noqa
        instruction=None,  # noqa
        name=base.LocalizedString({'EN': 'name'}),
        description=base.LocalizedString({'EN': 'desc'}),
        function=base.AnnotationFunction(
            inputs=(base.ObjectMeta(objects.Image, name='picture'),),
            outputs=(
                base.ObjectMeta(objects.Text, name='comment'),
                base.ClassMeta(Topic, name='topic'),
                base.ClassMeta(Review, name='review', available_labels=options_if),
            ),
        ),
    )
    experts_task_spec = spec.AnnotationTaskSpec(
        task_spec,
        'RU',
        project_params.Scenario.EXPERT_LABELING_OF_TASKS,
    )
    view, task_mapping = experts_task_spec.check.view, experts_task_spec.check.task_mapping
    assert task_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='picture'), (('url', 'picture'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='comment'), (('text', 'comment'),)),
            ObjectMapping(base.ClassMeta(Topic, name='topic'), (('value', 'topic'),)),
            ObjectMapping(base.ClassMeta(Review, name='review', available_labels=options_if), (('value', 'review'),)),
        ),
        output_mapping=(
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='eval',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the solution accurate?',
                                'KK': 'Тапсырманың шешімі дұрыс па?',
                                'RU': 'Решение верное?',
                                'TR': 'Görevin kararı doğru mu?',
                            }
                        )
                    ),
                ),
                (('ok', 'eval'),),
            ),
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='_ok',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the task (and the solution, if given) correct?',
                                'RU': 'Корректно ли задание (и решение, при наличии)?',
                            }
                        )
                    ),
                ),
                (('ok', '_ok'),),
            ),
            ObjectMapping(
                base.ObjectMeta(
                    objects.Text,
                    name='_comment',
                    title=base.Title(text=base.LocalizedString(lang_to_text={'EN': 'comment', 'RU': 'комментарий'})),
                    required=False,
                ),
                (('text', '_comment'),),
            ),
        ),
    )
    assert view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.ImageViewV1(url=tb.InputData(path='picture')),
                tb.TextViewV1(content=tb.InputData(path='comment')),
                tb.RadioGroupFieldV1(
                    data=tb.InternalData(path='topic', default=tb.InputData(path='topic')),
                    options=[
                        tb.GroupFieldOption(value='movie', label='movie'),
                        tb.GroupFieldOption(value='news', label='news'),
                    ],
                    disabled=True,
                ),
                tb.IfHelperV1(
                    condition=tb.EqualsConditionV1(
                        data=tb.InputData(path='topic'),
                        to='movie',
                    ),
                    then=tb.RadioGroupFieldV1(
                        data=tb.InternalData(path='review', default=tb.InputData(path='review')),
                        options=[
                            tb.GroupFieldOption(value='positive', label='positive'),
                            tb.GroupFieldOption(value='negative', label='negative'),
                            tb.GroupFieldOption(value='neutral', label='neutral'),
                        ],
                        disabled=True,
                    ),
                    else_=tb.IfHelperV1(
                        condition=tb.EqualsConditionV1(
                            data=tb.InputData(path='topic'),
                            to='news',
                        ),
                        then=tb.RadioGroupFieldV1(
                            data=tb.InternalData(path='review', default=tb.InputData(path='review')),
                            options=[
                                tb.GroupFieldOption(value='skeptical', label='skeptical'),
                                tb.GroupFieldOption(value='approving', label='approving'),
                            ],
                            disabled=True,
                        ),
                    ),
                ),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Решение верное?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='eval'),
                                options=[
                                    tb.GroupFieldOption(value=True, label='Да'),
                                    tb.GroupFieldOption(value=False, label='Нет'),
                                ],
                                disabled=False,
                                validation=tb.RequiredConditionV1(
                                    hint='Выберите один из вариантов', data=tb.OutputData(path='eval')
                                ),
                            ),
                        ]
                    )
                ),
                tb.MarkdownViewV1('---'),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Корректно ли задание (и решение, при наличии)?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='_ok'),
                                disabled=False,
                                options=[
                                    tb.GroupFieldOption(value=True, label='Да'),
                                    tb.GroupFieldOption(value=False, label='Нет'),
                                ],
                                validation=tb.RequiredConditionV1(
                                    hint='Выберите один из вариантов',
                                    data=tb.OutputData(path='_ok'),
                                ),
                            ),
                        ]
                    )
                ),
                tb.TextareaFieldV1(data=tb.OutputData(path='_comment'), placeholder='комментарий'),
            ]
        ),
        plugins=[
            tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData('eval'), payload=True)),
            tb.HotkeysPluginV1(key_2=tb.SetActionV1(data=tb.OutputData('eval'), payload=False)),
            tb.HotkeysPluginV1(key_3=tb.SetActionV1(data=tb.OutputData('_ok'), payload=True)),
            tb.HotkeysPluginV1(key_4=tb.SetActionV1(data=tb.OutputData('_ok'), payload=False)),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='scroll', task_width=700.0),
                notifications=[
                    tb.TextViewV1(content='Перед выполнением проверьте, что все медиа-элементы загрузились.'),
                    tb.TextViewV1(content='Если нет хотя бы одного медиа-элемента, перезагрузите страницу.'),
                ],
            ),
        ],
    )


def test_default_image_annotation():
    task_spec = spec.AnnotationTaskSpec(
        task_spec=base.TaskSpec(
            id=None,  # noqa
            instruction=None,  # noqa
            name=base.LocalizedString({'EN': 'name'}),
            description=base.LocalizedString({'EN': 'desc'}),
            function=base.AnnotationFunction(
                inputs=(objects.Image, objects.Text),
                outputs=(base.ImageAnnotation,),
            ),
        ),
        lang='EN',
    )

    image_annotation_view, image_annotation_mapping = task_spec.view, task_spec.task_mapping
    image_annotation_check_view, image_annotation_check_mapping = task_spec.check.view, task_spec.check.task_mapping

    assert image_annotation_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text'),)),
        ),
        output_mapping=(
            ObjectMapping(
                base.ObjectMeta(
                    base.ImageAnnotation,
                    name='image_annotation',
                ),
                (('data', 'image_annotation'),),
            ),
        ),
    )
    assert image_annotation_check_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text'),)),
            ObjectMapping(
                base.ObjectMeta(
                    base.ImageAnnotation,
                    name='image_annotation',
                ),
                (('data', 'image_annotation'),),
            ),
        ),
        output_mapping=(
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='eval',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the solution accurate?',
                                'KK': 'Тапсырманың шешімі дұрыс па?',
                                'RU': 'Решение верное?',
                                'TR': 'Görevin kararı doğru mu?',
                            }
                        )
                    ),
                ),
                (('ok', 'eval'),),
            ),
        ),
    )
    assert image_annotation_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text')),
                tb.ImageAnnotationFieldV1(
                    data=tb.OutputData(path='image_annotation'),
                    image=tb.InputData(path='image'),
                    validation=tb.conditions.RequiredConditionV1(
                        data=tb.OutputData(path='image_annotation'), hint='Select at least one area'
                    ),
                    full_height=True,
                ),
            ]
        ),
        plugins=[
            tb.ImageAnnotationHotkeysPluginV1(),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='pager'),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )
    assert image_annotation_check_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text')),
                tb.ImageAnnotationFieldV1(
                    data=tb.InternalData(path='image_annotation', default=tb.InputData(path='image_annotation')),
                    image=tb.InputData(path='image'),
                    disabled=True,
                    full_height=True,
                ),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Is the solution accurate?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='eval'),
                                options=[
                                    tb.GroupFieldOption(value=True, label='Yes'),
                                    tb.GroupFieldOption(value=False, label='No'),
                                ],
                                disabled=False,
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options', data=tb.OutputData(path='eval')
                                ),
                            ),
                        ]
                    )
                ),
            ]
        ),
        plugins=[
            tb.ImageAnnotationHotkeysPluginV1(),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='pager'),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )


def test_custom_image_annotation():
    class Area(base.Class):
        CAR = 'car'
        SIGN = 'sign'
        OTHER = 'other'

        @classmethod
        def labels(cls) -> Dict['Area', Dict[str, str]]:
            return {
                cls.CAR: {'EN': 'A car', 'RU': 'Машина'},
                cls.SIGN: {'EN': 'A sign', 'RU': 'Знак'},
                cls.OTHER: {'EN': 'Something else', 'RU': 'Другое'},
            }

    task_spec = spec.AnnotationTaskSpec(
        task_spec=base.TaskSpec(
            id=None,  # noqa
            instruction=None,  # noqa
            name=base.LocalizedString({'EN': 'name'}),
            description=base.LocalizedString({'EN': 'desc'}),
            function=base.AnnotationFunction(
                inputs=(
                    base.ObjectMeta(type=objects.Image),
                    base.ObjectMeta(objects.Text),
                ),
                outputs=(
                    base.ImageAnnotationMeta(
                        type=base.ImageAnnotation,
                        available_shapes={tb.fields.ImageAnnotationFieldV1.Shape.POLYGON},
                        labels=Area,
                    ),
                ),
            ),
        ),
        lang='EN',
    )

    image_annotation_view, image_annotation_mapping = task_spec.view, task_spec.task_mapping
    image_annotation_check_view, image_annotation_check_mapping = task_spec.check.view, task_spec.check.task_mapping

    assert image_annotation_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text'),)),
        ),
        output_mapping=(
            ObjectMapping(
                base.ImageAnnotationMeta(
                    base.ImageAnnotation,
                    name='image_annotation',
                    available_shapes={tb.fields.ImageAnnotationFieldV1.Shape.POLYGON},
                    labels=Area,
                ),
                (('data', 'image_annotation'),),
            ),
        ),
    )
    assert image_annotation_check_mapping == TaskMapping(
        input_mapping=(
            ObjectMapping(base.ObjectMeta(objects.Image, name='image'), (('url', 'image'),)),
            ObjectMapping(base.ObjectMeta(objects.Text, name='text'), (('text', 'text'),)),
            ObjectMapping(
                base.ImageAnnotationMeta(
                    base.ImageAnnotation,
                    name='image_annotation',
                    available_shapes={tb.fields.ImageAnnotationFieldV1.Shape.POLYGON},
                    labels=Area,
                ),
                (('data', 'image_annotation'),),
            ),
        ),
        output_mapping=(
            ObjectMapping(
                base.EvaluationMeta(
                    base.BinaryEvaluation,
                    name='eval',
                    title=base.Title(
                        text=base.LocalizedString(
                            lang_to_text={
                                'EN': 'Is the solution accurate?',
                                'KK': 'Тапсырманың шешімі дұрыс па?',
                                'RU': 'Решение верное?',
                                'TR': 'Görevin kararı doğru mu?',
                            }
                        )
                    ),
                ),
                (('ok', 'eval'),),
            ),
        ),
    )
    assert image_annotation_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text')),
                tb.ImageAnnotationFieldV1(
                    data=tb.OutputData(path='image_annotation'),
                    image=tb.InputData(path='image'),
                    labels=[
                        tb.ImageAnnotationFieldV1.Label(label=label, value=value)
                        for (label, value) in [('A car', 'car'), ('A sign', 'sign'), ('Something else', 'other')]
                    ],
                    shapes={tb.ImageAnnotationFieldV1.Shape.POLYGON: True},
                    validation=tb.conditions.RequiredConditionV1(
                        data=tb.OutputData(path='image_annotation'), hint='Select at least one area'
                    ),
                    full_height=True,
                ),
            ]
        ),
        plugins=[
            tb.ImageAnnotationHotkeysPluginV1(),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='pager'),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )
    assert image_annotation_check_view == toloka.project.TemplateBuilderViewSpec(
        view=tb.ListViewV1(
            items=[
                tb.TextViewV1(content=tb.InputData(path='text')),
                tb.ImageAnnotationFieldV1(
                    data=tb.InternalData(path='image_annotation', default=tb.InputData(path='image_annotation')),
                    image=tb.InputData(path='image'),
                    labels=[
                        tb.ImageAnnotationFieldV1.Label(label=label, value=value)
                        for (label, value) in [('A car', 'car'), ('A sign', 'sign'), ('Something else', 'other')]
                    ],
                    shapes={tb.ImageAnnotationFieldV1.Shape.POLYGON: True},
                    disabled=True,
                    full_height=True,
                ),
                tb.GroupViewV1(
                    content=tb.ListViewV1(
                        items=[
                            tb.TextViewV1(content='Is the solution accurate?'),
                            tb.RadioGroupFieldV1(
                                data=tb.OutputData(path='eval'),
                                options=[
                                    tb.GroupFieldOption(value=True, label='Yes'),
                                    tb.GroupFieldOption(value=False, label='No'),
                                ],
                                disabled=False,
                                validation=tb.RequiredConditionV1(
                                    hint='Choose one of the options', data=tb.OutputData(path='eval')
                                ),
                            ),
                        ]
                    )
                ),
            ]
        ),
        plugins=[
            tb.ImageAnnotationHotkeysPluginV1(),
            tb.TolokaPluginV1(
                layout=tb.TolokaPluginV1.TolokaPluginLayout(kind='pager'),
                notifications=[
                    tb.TextViewV1(content='Before performing a task, make sure that all media elements have loaded.'),
                    tb.TextViewV1(content='If at least one media element is missing, reload the page.'),
                ],
            ),
        ],
    )



from dataclasses import dataclass

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_

from ... import classification_loop
from .classification import TaskLabelsProbas
from .common import ProtobufSerializer, deserialize_one_of_field


@message
@dataclass
class StaticOverlap(ProtobufSerializer[classification_loop.StaticOverlap]):
    overlap: int = field(1)

    @staticmethod
    def serialize(obj: classification_loop.StaticOverlap) -> 'StaticOverlap':
        return StaticOverlap(overlap=obj.overlap)

    def deserialize(self) -> classification_loop.StaticOverlap:
        return classification_loop.StaticOverlap(overlap=self.overlap)


@message
@dataclass
class DynamicOverlap(ProtobufSerializer[classification_loop.DynamicOverlap]):
    min_overlap: int = field(1)
    max_overlap: int = field(2)
    confidence: OneOf_ = one_of(
        unified=part(float, 3),
        per_label=part(TaskLabelsProbas, 4),
    )

    @staticmethod
    def serialize(obj: classification_loop.DynamicOverlap) -> 'DynamicOverlap':
        overlap = DynamicOverlap(min_overlap=obj.min_overlap, max_overlap=obj.max_overlap)
        if isinstance(obj.confidence, float):
            overlap.confidence.unified = obj.confidence
        elif isinstance(obj.confidence, dict):
            overlap.confidence.per_label = TaskLabelsProbas.serialize(obj.confidence)
        else:
            raise ValueError(f'unexpected confidence type: {type(obj.confidence)}')
        return overlap

    def deserialize(self) -> classification_loop.DynamicOverlap:
        return classification_loop.DynamicOverlap(
            min_overlap=self.min_overlap,
            max_overlap=self.max_overlap,
            confidence=deserialize_one_of_field(self.confidence),
        )


@message
@dataclass
class Overlap(ProtobufSerializer[classification_loop.Overlap]):
    overlap: OneOf_ = one_of(
        static=part(StaticOverlap, 1),
        dynamic=part(DynamicOverlap, 2),
    )

    @staticmethod
    def serialize(obj: classification_loop.Overlap) -> 'Overlap':
        overlap = Overlap()
        if isinstance(obj, classification_loop.StaticOverlap):
            overlap.overlap.static = StaticOverlap.serialize(obj)
        elif isinstance(obj, classification_loop.DynamicOverlap):
            overlap.overlap.dynamic = DynamicOverlap.serialize(obj)
        else:
            raise ValueError(f'unexpected overlap type: {type(obj)}')
        return overlap

    def deserialize(self) -> classification_loop.Overlap:
        return deserialize_one_of_field(self.overlap)



from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from pure_protobuf.dataclasses_ import field, message

from .classification import AggregationAlgorithm
from .classification_loop import Overlap
from .common import ProtobufSerializer, func_registry, func_registry_reversed
from .control import Control
from .evaluation import AssignmentCheckSample
from .pricing import PricingStrategy
from .worker import HumanFilter, Model
from ...pricing import PoolPricingConfig
import crowdom.params as params


@message
@dataclass
class ExpertParams(ProtobufSerializer[params.ExpertParams]):
    task_duration_hint: timedelta = field(1)
    assignment_price: float = field(2)
    real_tasks_count: int = field(3)

    @staticmethod
    def serialize(obj: params.ExpertParams) -> 'ExpertParams':
        return ExpertParams(
            task_duration_hint=obj.task_duration_hint,
            assignment_price=obj.assignment_price,
            real_tasks_count=obj.real_tasks_count,
        )

    def deserialize(self) -> params.ExpertParams:
        return params.ExpertParams(
            task_duration_hint=self.task_duration_hint,
            pricing_config=PoolPricingConfig(
                assignment_price=self.assignment_price,
                real_tasks_count=self.real_tasks_count,
                control_tasks_count=0,
            ),
        )


@message
@dataclass
class Params(ExpertParams, ProtobufSerializer[params.Params]):
    worker_filter: HumanFilter = field(100)
    control_tasks_count: int = field(101)
    pricing_strategy: PricingStrategy = field(102)
    control: Control = field(103)
    overlap: Overlap = field(104)
    aggregation_algorithm: Optional[AggregationAlgorithm] = field(105, default=None)
    model: Optional[Model] = field(106, default=None)
    task_duration_function: Optional[str] = field(107, default=None)

    @staticmethod
    def serialize(obj: params.Params) -> 'Params':
        p = ExpertParams.serialize(obj)
        return Params(
            task_duration_hint=p.task_duration_hint,
            assignment_price=p.assignment_price,
            real_tasks_count=p.real_tasks_count,
            worker_filter=HumanFilter.serialize(obj.worker_filter),
            control_tasks_count=obj.control_tasks_count,
            pricing_strategy=PricingStrategy.serialize(obj.pricing_strategy),
            control=Control.serialize(obj.control),
            overlap=Overlap.serialize(obj.overlap),
            aggregation_algorithm=AggregationAlgorithm.serialize(obj.aggregation_algorithm)
            if obj.aggregation_algorithm
            else None,
            model=Model.serialize(obj.model) if obj.model else None,
            task_duration_function=func_registry[obj.task_duration_function] if obj.task_duration_function else None,
        )

    def deserialize(self) -> params.Params:
        p = super(Params, self).deserialize()
        return params.Params(
            task_duration_hint=p.task_duration_hint,
            pricing_config=PoolPricingConfig(
                assignment_price=p.pricing_config.assignment_price,
                real_tasks_count=p.pricing_config.real_tasks_count,
                control_tasks_count=self.control_tasks_count,
            ),
            overlap=Overlap.deserialize(self.overlap),
            control=Control.deserialize(self.control),
            worker_filter=HumanFilter.deserialize(self.worker_filter),
            aggregation_algorithm=AggregationAlgorithm.deserialize(self.aggregation_algorithm)
            if self.aggregation_algorithm
            else None,
            pricing_strategy=PricingStrategy.deserialize(self.pricing_strategy),
            task_duration_function=func_registry_reversed[self.task_duration_function]
            if self.task_duration_function
            else None,
        )


@message
@dataclass
class AnnotationParams(Params, ProtobufSerializer[params.AnnotationParams]):
    assignment_check_sample: Optional[AssignmentCheckSample] = field(200, default=None)

    @staticmethod
    def serialize(obj: params.AnnotationParams) -> 'AnnotationParams':
        p = Params.serialize(obj)
        return AnnotationParams(
            task_duration_hint=p.task_duration_hint,
            assignment_price=p.assignment_price,
            real_tasks_count=p.real_tasks_count,
            worker_filter=p.worker_filter,
            control_tasks_count=p.control_tasks_count,
            pricing_strategy=p.pricing_strategy,
            control=p.control,
            overlap=p.overlap,
            aggregation_algorithm=p.aggregation_algorithm,
            model=p.model,
            task_duration_function=p.task_duration_function,
            assignment_check_sample=AssignmentCheckSample.serialize(obj.assignment_check_sample)
            if obj.assignment_check_sample
            else None,
        )

    def deserialize(self) -> params.AnnotationParams:
        p = super(AnnotationParams, self).deserialize()
        annotation_params = params.AnnotationParams(
            task_duration_hint=p.task_duration_hint,
            pricing_config=p.pricing_config,
            overlap=p.overlap,
            control=p.control,
            worker_filter=p.worker_filter,
            aggregation_algorithm=p.aggregation_algorithm,
            pricing_strategy=p.pricing_strategy,
            task_duration_function=p.task_duration_function,
            assignment_check_sample=self.assignment_check_sample.deserialize()
            if self.assignment_check_sample
            else None,
        )
        if self.model:
            annotation_params.model = Model.deserialize(self.model)
        return annotation_params



from datetime import timedelta
from mock import patch
from pytest import approx, fixture
from toloka.client.assignment import Assignment
from toloka.client.user_restriction import UserRestriction
from typing import List

from crowdom import (
    base,
    classification,
    classification_loop,
    control,
    pricing,
    params,
    objects,
    task_spec as spec,
    worker,
)
from crowdom.control import (
    AssignmentAccuracyPredicate,
    SetAssignmentStatus,
    BlockUser,
    PredicateExpression,
    ComparisonType,
    Rule,
    BooleanOperator,
    AssignmentDurationPredicate,
)

from . import lib


class TestGetValues:
    def test_assignment_price(self):
        assert params.AssignmentPrice.get_values(
            task_duration_hint=timedelta(seconds=5),
            task_count=1,
            control_task_count=0,
        ) == (0.005, [0.005, 0.208], [0.005, 0.005])

        assert params.AssignmentPrice.get_values(
            task_duration_hint=timedelta(seconds=30),
            task_count=30,
            control_task_count=30,
        ) == (0.3, [0.005, 1.25], [0.15, 0.45])

    def test_annotation_assignment_price(self):
        assert params.AnnotationAssignmentPrice.get_values(
            task_duration_hint=timedelta(seconds=5),
            task_count=1,
        ) == (0.005, [0.005, 0.104], [0.005, 0.005])

        assert params.AnnotationAssignmentPrice.get_values(
            task_duration_hint=timedelta(seconds=30),
            task_count=30,
        ) == (0.15, [0.005, 0.625], [0.075, 0.225])

    def test_control_task_count(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.NO_CONTROL, (0, [0, 10], [1, 4], True)),
            (params.QualityPresetType.MILD, (2, [0, 10], [1, 4], True)),
            (params.QualityPresetType.MODERATE, (3, [0, 10], [1, 4], True)),
            (params.QualityPresetType.STRICT, (6, [0, 10], [1, 4], True)),
        ]:
            assert (
                params.ControlTaskCount.get_values(
                    task_count=15,
                    quality_preset=quality_preset.value,
                )
                == params.ControlTaskCount.get_values(
                    task_count=15,
                    quality_preset=quality_preset.value,
                    control_task_count=5,
                )
                == expected_values
            )

        assert params.ControlTaskCount.get_values(
            task_count=15,
            quality_preset=params.QualityPresetType.CUSTOM.value,
            control_task_count=5,
        ) == (5, [0, 15], [1, 4], False)

    def test_control_tasks_for_accept(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.NO_CONTROL, (0, True, [0, 12], [3, 8])),
            (params.QualityPresetType.MILD, (5, True, [0, 12], [3, 8])),
            (params.QualityPresetType.MODERATE, (6, True, [0, 12], [3, 8])),
            (params.QualityPresetType.STRICT, (7, True, [0, 12], [3, 8])),
        ]:
            assert (
                params.ControlTasksForAccept.get_values(
                    num_classes=3,
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                )
                == params.ControlTasksForAccept.get_values(
                    num_classes=3,
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                    control_tasks_for_accept=10,
                )
                == expected_values
            )

        assert params.ControlTasksForAccept.get_values(
            num_classes=3,
            quality_preset=params.QualityPresetType.CUSTOM.value,
            control_task_count=12,
            control_tasks_for_accept=10,
        ) == (10, False, [0, 12], [3, 8])

    def test_checked_tasks_for_accept(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.MILD, (8, True, [0, 12], [3, 8])),
            (params.QualityPresetType.MODERATE, (9, True, [0, 12], [3, 8])),
            (params.QualityPresetType.STRICT, (11, True, [0, 12], [3, 8])),
        ]:
            assert (
                params.CheckedTasksForAccept.get_values(
                    num_classes=3,
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                )
                == params.CheckedTasksForAccept.get_values(
                    num_classes=3,
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                    control_tasks_for_accept=10,
                )
                == expected_values
            )

        assert params.CheckedTasksForAccept.get_values(
            num_classes=3,
            quality_preset=params.QualityPresetType.CUSTOM.value,
            control_task_count=12,
            control_tasks_for_accept=10,
        ) == (10, False, [0, 12], [3, 8])

    def test_control_tasks_for_block(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.NO_CONTROL, (0, True, [0, 12], [0, 4])),
            (params.QualityPresetType.MILD, (2, True, [0, 12], [0, 4])),
            (params.QualityPresetType.MODERATE, (4, True, [0, 12], [0, 4])),
            (params.QualityPresetType.STRICT, (6, True, [0, 12], [0, 4])),
        ]:
            assert (
                params.ControlTasksForBlock.get_values(
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                )
                == params.ControlTasksForBlock.get_values(
                    quality_preset=quality_preset.value,
                    control_task_count=12,
                    control_tasks_for_block=10,
                )
                == expected_values
            )

        assert params.ControlTasksForBlock.get_values(
            quality_preset=params.QualityPresetType.CUSTOM.value,
            control_task_count=12,
            control_tasks_for_block=10,
        ) == (10, False, [0, 12], [0, 4])

    def test_overlap(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.NO_CONTROL, ('1', True)),
            (params.QualityPresetType.MILD, ('2-3', True)),
            (params.QualityPresetType.MODERATE, ('3', True)),
            (params.QualityPresetType.STRICT, ('3-5', True)),
        ]:
            assert (
                params.Overlap.get_values(
                    quality_preset=quality_preset.value,
                )
                == params.Overlap.get_values(
                    quality_preset=quality_preset.value,
                    overlap='1-5',
                )
                == expected_values
            )

        assert params.Overlap.get_values(
            quality_preset=params.QualityPresetType.CUSTOM.value,
            overlap='1-5',
        ) == ('1-5', False)

    def test_max_attempts(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.MILD, ('2', True)),
            (params.QualityPresetType.MODERATE, ('3', True)),
            (params.QualityPresetType.STRICT, ('5', True)),
        ]:
            assert (
                params.MaxAttempts.get_values(
                    quality_preset=quality_preset.value,
                )
                == params.MaxAttempts.get_values(
                    quality_preset=quality_preset.value,
                    max_attempts='5',
                )
                == expected_values
            )

        assert params.MaxAttempts.get_values(
            quality_preset=params.QualityPresetType.CUSTOM.value,
            max_attempts='5',
        ) == ('5', False)

    def test_speed_control(self):
        for quality_preset, expected_values in [
            (params.QualityPresetType.NO_CONTROL, ([], True)),
            (
                params.QualityPresetType.MILD,
                ([params.BlockTimePicker(0.01, '1d', True), params.BlockTimePicker(0.05, '2h', False)], True),
            ),
            (
                params.QualityPresetType.MODERATE,
                ([params.BlockTimePicker(0.1, '1d', True), params.BlockTimePicker(0.3, '2h', False)], True),
            ),
            (
                params.QualityPresetType.STRICT,
                ([params.BlockTimePicker(0.3, '1d', True), params.BlockTimePicker(0.5, '2h', False)], True),
            ),
        ]:
            assert (
                params.SpeedControlList.get_values(
                    quality_preset=quality_preset.value,
                )
                == params.SpeedControlList.get_values(
                    quality_preset=quality_preset.value,
                    items=[params.BlockTimePicker()],
                )
                == expected_values
            )

        items = [params.BlockTimePicker(0.5, '1d', False), params.BlockTimePicker(0.35, '2d', True)]
        assert params.SpeedControlList.get_values(
            quality_preset=params.QualityPresetType.CUSTOM.value,
            items=items,
        ) == (items, False)

    def test_check_sample(self):
        for quality_preset in [
            params.QualityPresetType.MILD,
            params.QualityPresetType.MODERATE,
            params.QualityPresetType.STRICT,
        ]:
            assert (
                params.CheckSample.get_values(
                    quality_preset=quality_preset.value,
                )
                == params.CheckSample.get_values(
                    quality_preset=quality_preset.value,
                    check_sample=False,
                )
                == (False, True)
            )

        assert params.CheckSample.get_values(
            quality_preset=params.QualityPresetType.CUSTOM.value,
            check_sample=True,
        ) == (True, False)

    def test_max_tasks_to_check(self):
        for quality_preset in [
            params.QualityPresetType.MILD,
            params.QualityPresetType.MODERATE,
            params.QualityPresetType.STRICT,
        ]:
            assert (
                params.MaxTasksToCheck.get_values(
                    task_count=15,
                    quality_preset=quality_preset.value,
                    check_sample=True,
                )
                == params.MaxTasksToCheck.get_values(
                    task_count=15,
                    quality_preset=quality_preset.value,
                    check_sample=True,
                    max_tasks_to_check=10,
                )
                == (15, True, [1, 15], [15])
            )

        assert params.MaxTasksToCheck.get_values(
            task_count=15,
            quality_preset=params.QualityPresetType.CUSTOM.value,
            check_sample=True,
            max_tasks_to_check=5,
        ) == (5, False, [1, 15], [15])

    def test_accuracy_thresold(self):
        for quality_preset in [
            params.QualityPresetType.MILD,
            params.QualityPresetType.MODERATE,
            params.QualityPresetType.STRICT,
        ]:
            assert params.AccuracyThreshold.get_values(
                task_count=15,
                quality_preset=quality_preset.value,
                check_sample=True,
                max_tasks_to_check=10,
            ) == (10, True, [1, 10], [8, 10])

        assert params.AccuracyThreshold.get_values(
            task_count=15,
            quality_preset=params.QualityPresetType.CUSTOM.value,
            check_sample=True,
            max_tasks_to_check=10,
            accuracy_threshold=5,
        ) == (5, False, [1, 10], [8, 10])


@fixture
def task_duration_hint() -> timedelta:
    return timedelta(seconds=10)


@fixture
def initial_rules(task_duration_hint: timedelta) -> List[control.Rule]:
    return [
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=1.0,
                comparison=ComparisonType('>='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.ACCEPTED),
        ),
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=1.0,
                comparison=ComparisonType('<'),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.1,
                comparison=ComparisonType('<='),
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, <= 0.1',
                duration=timedelta(days=1),
            ),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.1,
                comparison=ComparisonType('<='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.1,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.1 < time <= 0.3',
                duration=timedelta(hours=2),
            ),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentAccuracyPredicate(
                        threshold=0.0,
                        comparison=ComparisonType('>='),
                    ),
                    AssignmentAccuracyPredicate(
                        threshold=0.5,
                        comparison=ComparisonType('<'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Control tasks: [0, 1) done correctly',
                duration=timedelta(hours=1),
            ),
        ),
    ]


@fixture
def initial_annotation_rules(task_duration_hint: timedelta) -> List[control.Rule]:
    return [
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=7 / 9,
                comparison=ComparisonType('>='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.ACCEPTED),
        ),
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=7 / 9,
                comparison=ComparisonType('<'),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.1,
                comparison=ComparisonType('<='),
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, <= 0.1',
                duration=timedelta(days=1),
            ),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.1,
                comparison=ComparisonType('<='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.1,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.1 < time <= 0.3',
                duration=timedelta(hours=2),
            ),
        ),
    ]


@fixture
def initial_params(task_duration_hint: timedelta, initial_rules: List[control.Rule]) -> params.Params:
    return params.Params(
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.UNITED_STATES, worker.RegionCodes.UNITED_KINGDOM},
                )
            ],
            training_score=None,
        ),
        task_duration_hint=task_duration_hint,
        pricing_config=pricing.PoolPricingConfig(
            assignment_price=0.018,
            real_tasks_count=9,
            control_tasks_count=2,
        ),
        overlap=classification_loop.StaticOverlap(3),
        control=control.Control(
            rules=initial_rules,
        ),
        aggregation_algorithm=classification.AggregationAlgorithm.DAWID_SKENE,
    )


@fixture
def initial_annotation_params(
    task_duration_hint: timedelta, initial_annotation_rules: List[control.Rule]
) -> params.AnnotationParams:
    return params.AnnotationParams(
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.UNITED_STATES, worker.RegionCodes.UNITED_KINGDOM},
                )
            ],
            training_score=None,
        ),
        task_duration_hint=task_duration_hint,
        pricing_config=pricing.PoolPricingConfig(
            assignment_price=0.014,
            real_tasks_count=9,
            control_tasks_count=0,
        ),
        overlap=classification_loop.DynamicOverlap(1, 3, 0.85),
        control=control.Control(
            rules=initial_annotation_rules,
        ),
        aggregation_algorithm=None,
    )


@fixture
def changed_rules(task_duration_hint: timedelta) -> List[control.Rule]:
    return [
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=7 / 12,
                comparison=ComparisonType('>='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.ACCEPTED),
        ),
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=7 / 12,
                comparison=ComparisonType('<'),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.3,
                comparison=ComparisonType('<='),
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, <= 0.3',
                duration=timedelta(days=1),
            ),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.3,
                comparison=ComparisonType('<='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.5,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.3 < time <= 0.5',
                duration=timedelta(hours=2),
            ),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentAccuracyPredicate(
                        threshold=0.0,
                        comparison=ComparisonType('>='),
                    ),
                    AssignmentAccuracyPredicate(
                        threshold=0.5,
                        comparison=ComparisonType('<'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Control tasks: [0, 6) done correctly',
                duration=timedelta(hours=1),
            ),
        ),
    ]


@fixture
def changed_params(task_duration_hint: timedelta, changed_rules: List[control.Rule]) -> params.Params:
    return params.Params(
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.UNITED_STATES, worker.RegionCodes.UNITED_KINGDOM},
                )
            ],
            training_score=None,
        ),
        task_duration_hint=task_duration_hint,
        pricing_config=pricing.PoolPricingConfig(
            assignment_price=0.07,
            real_tasks_count=30,
            control_tasks_count=12,
        ),
        overlap=classification_loop.DynamicOverlap(3, 5, 0.85),
        control=control.Control(
            rules=changed_rules,
        ),
        aggregation_algorithm=classification.AggregationAlgorithm.DAWID_SKENE,
    )


@fixture
def changed_annotation_rules(task_duration_hint: timedelta) -> List[control.Rule]:
    return [
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=0.9,
                comparison=ComparisonType('>='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.ACCEPTED),
        ),
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=0.9,
                comparison=ComparisonType('<'),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.3,
                comparison=ComparisonType('<='),
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, <= 0.3',
                duration=timedelta(days=1),
            ),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.3,
                comparison=ComparisonType('<='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.5,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.3 < time <= 0.5',
                duration=timedelta(hours=2),
            ),
        ),
    ]


@fixture
def changed_annotation_params(
    task_duration_hint: timedelta, changed_annotation_rules: List[control.Rule]
) -> params.AnnotationParams:
    return params.AnnotationParams(
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.UNITED_STATES, worker.RegionCodes.UNITED_KINGDOM},
                )
            ],
            training_score=None,
        ),
        task_duration_hint=task_duration_hint,
        pricing_config=pricing.PoolPricingConfig(
            assignment_price=0.05,
            real_tasks_count=30,
            control_tasks_count=0,
        ),
        overlap=classification_loop.DynamicOverlap(1, 5, 0.85),
        control=control.Control(
            rules=changed_annotation_rules,
        ),
        aggregation_algorithm=None,
        assignment_check_sample=None,
    )


@fixture
def final_rules(task_duration_hint: timedelta) -> List[control.Rule]:
    return [
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=1.0,
                comparison=ComparisonType('>='),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.ACCEPTED),
        ),
        Rule(
            predicate=AssignmentAccuracyPredicate(
                threshold=1.0,
                comparison=ComparisonType('<'),
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=AssignmentDurationPredicate(
                threshold=0.05,
                comparison=ComparisonType('<='),
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, <= 0.05',
                duration=timedelta(minutes=30),
            ),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.05,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.05 < time <= 0.3',
                duration=timedelta(days=1),
            ),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.05,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=SetAssignmentStatus(status=Assignment.Status.REJECTED),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentDurationPredicate(
                        threshold=0.5,
                        comparison=ComparisonType('<='),
                    ),
                    AssignmentDurationPredicate(
                        threshold=0.3,
                        comparison=ComparisonType('>'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Fast submits, 0.3 < time <= 0.5',
                duration=timedelta(hours=2),
            ),
        ),
        Rule(
            predicate=PredicateExpression(
                boolean_operator=BooleanOperator.AND,
                predicates=[
                    AssignmentAccuracyPredicate(
                        threshold=0.0,
                        comparison=ComparisonType('>='),
                    ),
                    AssignmentAccuracyPredicate(
                        threshold=1.0,
                        comparison=ComparisonType('<'),
                    ),
                ],
            ),
            action=BlockUser(
                scope=UserRestriction.Scope.POOL,
                private_comment='Control tasks: [0, 2) done correctly',
                duration=timedelta(hours=1),
            ),
        ),
    ]


@fixture
def final_params(task_duration_hint: timedelta, final_rules: List[control.Rule]) -> params.Params:
    return params.Params(
        worker_filter=worker.WorkerFilter(
            filters=[
                worker.WorkerFilter.Params(
                    langs={worker.LanguageRequirement(lang='EN', verified=True)},
                    regions={worker.RegionCodes.UNITED_STATES, worker.RegionCodes.UNITED_KINGDOM},
                )
            ],
            training_score=None,
        ),
        task_duration_hint=task_duration_hint,
        pricing_config=pricing.PoolPricingConfig(
            assignment_price=0.04,
            real_tasks_count=30,
            control_tasks_count=2,
        ),
        overlap=classification_loop.DynamicOverlap(3, 5, 0.85),
        control=control.Control(
            rules=final_rules,
        ),
        aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
    )


@patch('crowdom.params.characteristics.ITERATIONS_COUNT', 200)
class TestEventLoop:
    def test_loop(
        self,
        task_duration_hint: timedelta,
        initial_params: params.Params,
        changed_params: params.Params,
        final_params: params.Params,
    ):
        function = base.ClassificationFunction(inputs=(objects.Image,), cls=lib.ImageClass)
        instruction = {'EN': 'Identify the animal in the photo', 'RU': 'Определите, какое животное на фотографии'}
        task_spec = base.TaskSpec(
            id='dogs&cats',
            function=function,
            name={'EN': 'Cat or dog', 'RU': 'Кошка или собака'},
            description={'EN': 'Identification of animals in photos', 'RU': 'Определение животных на изображениях'},
            instruction=instruction,
        )
        task_spec_en = spec.PreparedTaskSpec(task_spec, 'EN')

        toloka_client = lib.TolokaClientCallRecorderStub()
        loop = params.get_loop(
            task_spec_en,
            task_duration_hint,
            toloka_client,  # noqa
        )
        loop.ui_initialized = True

        # generated default
        assert loop.get_params() == initial_params
        assert [p.value for p in loop.stats] == approx([0.87, 0.69, 0.30], abs=0.05)
        assert loop.characteristics.value[:5] == [
            'Robustness: 0.89, recommended range - [0.6, 0.9]',
            'Assignment duration: 0:01:50, recommended range - [30s, 5m]',
            'Assignment commission, usd: 0.005',
            'Expected worker earnings per hour: 0.589\\$',
            'Expected price per 1000 objects: 7.86\\$',
        ]
        # we can't check 'Expected loop iterations: ?'
        assert loop.characteristics.value[-1] == 'Task attractiveness index: 4 / 7'

        loop.start(params.UIEvent(params.ParamName.TASK_COUNT.ui, {params.ParamName.TASK_COUNT: params.Update(30)}))
        loop.start(
            params.UIEvent(
                params.ParamName.QUALITY_PRESET.ui, {params.ParamName.QUALITY_PRESET: params.Update('Strict')}
            )
        )

        # check
        assert loop.get_params() == changed_params
        assert [p.value for p in loop.stats] == approx([0.93, 0.76, 0.58], abs=0.05)
        assert loop.characteristics.value[:5] == [
            'Robustness: 0.93, recommended range - [0.6, 0.9]',
            'Assignment duration: 0:07:00, recommended range - [30s, 5m]',
            'Assignment commission, usd: 0.021',
            'Expected worker earnings per hour: 0.600\\$',
            'Expected price per 1000 objects: 9.28\\$ ... 15.47\\$',
        ]
        # can't check:
        # 'Expected loop iterations: ?',
        # 'Expected average overlap: ?',
        # 'Expected average price per 1000 objects: ?',
        assert loop.characteristics.value[-1] == 'Task attractiveness index: 3 / 7'

        # some more changes
        loop.start(
            params.UIEvent(
                params.ParamName.QUALITY_PRESET.ui, {params.ParamName.QUALITY_PRESET: params.Update('Custom')}
            )
        )
        loop.start(
            params.SpeedControlEvent(
                params.ParamName.SPEED_CONTROL.ui, {params.SpeedControlChangeType.ADD: params.SpeedControlUpdate('')}
            )
        )
        loop.start(
            params.UIEvent(
                params.ParamName.AGGREGATION_ALGORITHM.ui, {params.ParamName.AGGREGATION_ALGORITHM: params.Update('ML')}
            )
        )
        loop.start(
            params.UIEvent(
                params.ParamName.CONTROL_TASK_COUNT.ui, {params.ParamName.CONTROL_TASK_COUNT: params.Update(2)}
            )
        )
        loop.start(
            params.UIEvent(
                params.ParamName.ASSIGNMENT_PRICE.ui, {params.ParamName.ASSIGNMENT_PRICE: params.Update(0.04)}
            )
        )

        # final check
        assert loop.get_params() == final_params
        assert [p.value for p in loop.stats] == approx([0.87, 0.80, 0.21], abs=0.05)
        assert loop.characteristics.value[:5] == [
            'Robustness: 0.89, recommended range - [0.6, 0.9]',
            'Assignment duration: 0:05:20, recommended range - [30s, 5m]',
            'Assignment commission, usd: 0.012',
            'Expected worker earnings per hour: 0.450\\$',
            'Expected price per 1000 objects: 5.30\\$ ... 8.84\\$',
        ]
        # can't check:
        # 'Expected loop iterations: ?',
        # 'Expected average overlap: ?',
        # 'Expected average price per 1000 objects: ?',
        assert loop.characteristics.value[-1] == 'Task attractiveness index: 2 / 7'

    def test_annotation_loop(
        self,
        task_duration_hint: timedelta,
        initial_annotation_params: params.AnnotationParams,
        changed_annotation_params: params.AnnotationParams,
    ):
        function = base.AnnotationFunction(inputs=(objects.Audio,), outputs=(objects.Text,))
        instruction = {'EN': 'Transcribe given audio', 'RU': 'Запишите текст с аудио'}
        task_spec = base.TaskSpec(
            id='dogs&cats',
            function=function,
            name={'EN': 'Audio transcription', 'RU': 'Расшифровка речи'},
            description={'EN': 'Transcribe given audio', 'RU': 'Запишите текст с аудио'},
            instruction=instruction,
        )
        task_spec_en = spec.AnnotationTaskSpec(task_spec, 'EN')

        loop = params.get_loop_for_annotation(
            task_spec_en,
            task_duration_hint,
            toloka_client=None,  # noqa
        )

        loop.ui_initialized = True

        # generated default
        assert loop.get_params() == initial_annotation_params

        assert [p.value for p in loop.stats] == approx([0.63, 0.56, 0.40], abs=0.05)
        assert loop.characteristics.value[:3] == [
            'Assignment duration: 0:01:30, recommended range - [30s, 5m]',
            'Assignment commission, usd: 0.004',
            'Expected worker earnings per hour: 0.560\\$',
        ]
        # we can't check 'Expected loop iterations: ?'
        # and 'Expected average overlap: ?'
        assert loop.characteristics.value[-1] == 'Task attractiveness index: 4 / 6'

        loop.start(params.UIEvent(params.ParamName.TASK_COUNT.ui, {params.ParamName.TASK_COUNT: params.Update(30)}))
        loop.start(
            params.UIEvent(
                params.ParamName.QUALITY_PRESET.ui, {params.ParamName.QUALITY_PRESET: params.Update('Strict')}
            )
        )

        # check
        assert loop.get_params() == changed_annotation_params
        assert [p.value for p in loop.stats] == approx([0.89, 0.96, 0.03], abs=0.05)
        assert loop.characteristics.value[:2] == [
            'Assignment duration: 0:05:00, recommended range - [30s, 5m]',
            'Assignment commission, usd: 0.015',
        ]

        assert loop.characteristics.value[-1] == 'Task attractiveness index: 2 / 6'
